from typing import Callable, TypeVar, Tuple

from amino import Either, do, Do, _, IO, List

from ribosome.compute.tpe import analyse_prog_tpe, prog_type
from ribosome.compute.wrap import prog_wrappers
from ribosome.compute.output import (ProgOutput, ProgOutputUnit, ProgOutputResult, ProgOutputIO,
                                     ProgScalarIO, ProgGatherIOs, ProgScalarSubprocess, ProgGatherSubprocesses, Echo,
                                     ProgIOEcho)
from ribosome.nvim.io.state import NS
from ribosome.request.args import ParamsSpec
from ribosome.compute.wrap_data import ProgWrappers
from ribosome.config.settings import Settings
from ribosome.config.basic_config import NoData
from ribosome.compute.ribosome import Ribosome
from ribosome.compute.tpe_data import ribo_state_prog
from ribosome.compute.prog import Prog
from ribosome.compute.program import Program, ProgramBlock, ProgramCompose
from ribosome.process import Subprocess

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
M = TypeVar('M')
P = TypeVar('P')
R = TypeVar('R')
S = TypeVar('S', bound=Settings)
CC = TypeVar('CC')
C = TypeVar('C')
PIO = TypeVar('PIO')


def prog_type_error(func: Callable[[P], NS[R, A]], error: str) -> None:
    raise Exception(f'program `{func.__name__}` has invalid type: {error}')


def program_from_data(
        func: Callable[[P], NS[R, A]],
        params_spec: ParamsSpec,
        wrappers: ProgWrappers,
        interpreter: ProgOutput[A, B],
) -> Program[A]:
    return Program(func.__name__, ProgramBlock(func, wrappers, interpreter), params_spec)


def prog_state(func: Callable[[P], NS[R, A]], interpreter: ProgOutput[A, B]) -> Program:
    params_spec = ParamsSpec.from_function(func)
    wrappers = prog_type(func, params_spec).value_or(lambda err: prog_type_error(func, err))
    return Program(func.__name__, ProgramBlock(func, wrappers, interpreter), params_spec)


def func_state_data(f: Callable[..., NS[R, A]]) -> Tuple[ParamsSpec, ProgWrappers]:
    params_spec = ParamsSpec.from_function(f)
    @do(Either[str, ProgWrappers])
    def wrappers() -> Do:
        comp_type = yield params_spec.state_type.to_either('no state type')
        state_prog = ribo_state_prog(comp_type)
        yield prog_wrappers.match(state_prog)
    wrappers = wrappers().value_or(lambda err: prog_type_error(f, err))
    return params_spec, wrappers


class prog_io:

    def __new__(cls, func: Callable[[P], NS[D, IO[A]]]) -> Program[A]:
        return prog_state(func, ProgOutputIO(ProgScalarIO()))

    @staticmethod
    def gather(func: Callable[[P], NS[D, List[IO[A]]]]) -> Program[A]:
        return prog_state(func, ProgOutputIO(ProgGatherIOs()))


class prog_subproc:

    def __new__(cls, func: Callable[[P], NS[D, Subprocess[A]]]) -> Program[A]:
        return prog_state(func, ProgOutputIO(ProgScalarSubprocess()))

    @staticmethod
    def gather(func: Callable[[P], NS[D, List[Subprocess[A]]]]) -> Program[A]:
        return prog_state(func, ProgOutputIO(ProgGatherSubprocesses()))


class prog:

    def __new__(self, func: Callable[[P], NS[D, A]]) -> Program[A]:
        return prog_state(func, ProgOutputResult())

    @staticmethod
    def unit(func: Callable[[P], NS[D, None]]) -> Program[None]:
        return prog_state(func, ProgOutputUnit())

    @staticmethod
    def result(func: Callable[[P], NS[D, A]]) -> Program[A]:
        return prog_state(func, ProgOutputResult())

    @staticmethod
    def strict(func: Callable[[P], A]) -> Program[A]:
        def wrap(*p: P) -> NS[NoData, A]:
            return NS.pure(func(*p))
        return prog_state(wrap, ProgOutputResult())

    @staticmethod
    def do(func: Callable[[P], Do]) -> Program:
        f = do(Prog[A])(func)
        params_spec = ParamsSpec.from_function(f)
        return Program(func.__name__, ProgramCompose(f), params_spec)

    io = prog_io
    subproc = prog_subproc

    @staticmethod
    def comp(f: Callable[[P], NS[C, A]]) -> Program[A]:
        params_spec, wrappers = func_state_data(f)
        @do(NS[Ribosome[S, D, CC, C], A])
        def zoomed(*p: P) -> Do:
            comp_lens = yield NS.inspect(_.comp_lens)
            yield f(*p).zoom(comp_lens)
        return program_from_data(zoomed, params_spec, wrappers, ProgOutputResult())

    @staticmethod
    def echo(f: Callable[[P], NS[D, Echo]]) -> Program[None]:
        return prog_state(f, ProgOutputIO(ProgIOEcho()))


__all__ = ('prog',)



from typing import TypeVar, Any, Type, Callable

from amino import Try, curried, Lists, Either, do, Do

from msgpack import ExtType
from amino.util.string import decode

from ribosome.nvim.io.compute import NvimIO, NvimIOPure, NvimIORequest
from ribosome.nvim.io.trace import NvimIOException
from ribosome.nvim.io.compute import NvimIOError
from ribosome.nvim.io.cons import nvimio_recover_fatal, nvimio_from_either

A = TypeVar('A')


def nvim_error_msg(exc: Exception) -> str:
    return Try(lambda: decode(exc.args[0])) | str(exc)


def nvim_request_error(name: str, args: tuple, desc: str, error: Any) -> NvimIO[A]:
    msg = (nvim_error_msg(error.cause) if isinstance(error, NvimIOException) else str(error))
    return NvimIOError(f'{desc} in nvim request `{name}({Lists.wrap(args).join_comma})`: {msg}')


@decode.register(ExtType)
def decode_ext_type(a: ExtType) -> ExtType:
    return a


@do(NvimIO[Either[str, A]])
def nvim_nonfatal_request(name: str, *args: Any) -> Do:
    request = NvimIORequest(name, Lists.wrap(args))
    value = yield nvimio_recover_fatal(request, lambda a: nvim_request_error(name, args, 'fatal error', a))
    return value / decode


@do(NvimIO[A])
def nvim_request(name: str, *args: Any) -> Do:
    result = yield nvim_nonfatal_request(name, *args)
    yield nvimio_from_either(result)


@do(NvimIO[A])
def typechecked_request(name: str, tpe: Type[A], *args: Any) -> Do:
    raw = yield nvim_request(name, *args)
    yield (
        NvimIOPure(raw)
        if isinstance(raw, tpe) else
        NvimIOError(f'invalid result type of request {name}{args}: {raw}')
    )


@do(NvimIO[A])
def data_cons_request(name: str, cons: Callable[[Any], Either[str, A]], *args: Any) -> Do:
    raw = yield nvim_request(name, *args)
    yield nvimio_from_either(cons(raw))


@do(NvimIO[Either[str, A]])
def data_cons_request_nonfatal(name: str, cons: Callable[[Either[str, Any]], Either[str, A]], *args: Any) -> Do:
    raw = yield nvim_nonfatal_request(name, *args)
    return cons(raw)


__all__ = ('nvim_nonfatal_request', 'nvim_request', 'typechecked_request', 'data_cons_request',
           'data_cons_request_nonfatal')



from typing import Generic, Any, TypeVar, Callable

from amino import List, Lists, Map, _, Boolean, __
from amino.func import flip
from amino.util.string import ToStr, camelcase
from amino.dat import ADT
from amino.boolean import false

from ribosome.request.rpc import RpcHandlerSpec
from ribosome.request.handler.prefix import PrefixStyle, Short
from ribosome.request.handler.method import RpcMethod, CmdMethod, FunctionMethod, AutocmdMethod
from ribosome.nvim.io.compute import NvimIO

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
P = TypeVar('P')
RHS = TypeVar('RHS', bound=RpcHandlerSpec)
Meth = TypeVar('Meth', bound=RpcMethod)


def program_rpc_options(program: P) -> Map[str, str]:
    return Map(nargs=program.params_spec.nargs.for_vim)


class RequestHandler(Generic[Meth, P], ADT['RequestHandler[Meth, P]']):

    def __init__(
            self,
            method: Meth,
            program: P,
            name: str,
            prefix: PrefixStyle,
            sync: Boolean,
            json: Boolean,
            extra_options: Map[str, Any],
    ) -> None:
        self.method = method
        self.program = program
        self.name = name
        self.prefix = prefix
        self.sync = sync
        self.json = json
        self.extra_options = extra_options

    @staticmethod
    def trans_cmd(prog: P) -> 'RequestHandlerBuilder[CmdMethod]':
        return RequestHandlerBuilder(CmdMethod(), prog)

    @staticmethod
    def trans_function(prog: P) -> 'RequestHandlerBuilder[FunctionMethod]':
        return RequestHandlerBuilder(FunctionMethod(), prog)

    @staticmethod
    def trans_autocmd(prog: P) -> 'RequestHandlerBuilder[AutocmdMethod]':
        return RequestHandlerBuilder(AutocmdMethod(), prog)

    @property
    def options(self) -> Map[str, Any]:
        return self.extra_options ** Map(sync=self.sync) ** program_rpc_options(self.program)

    @property
    def method_str(self) -> str:
        return f'{self.method.method}:{self.name}'

    def vim_cmd_name(self, name: str, prefix: str) -> str:
        pre = name if self.prefix.full else prefix if self.prefix.short else ''
        return f'{camelcase(pre)}{camelcase(self.name)}'


class RequestHandlerBuilder(Generic[Meth]):

    def __init__(self, method: Meth, program: P) -> None:
        self.method = method
        self.program = program

    @property
    def sync_default(self) -> Boolean:
        return Boolean.isinstance(self.method, FunctionMethod)

    def __call__(
            self,
            name: str=None,
            prefix: PrefixStyle=Short(),
            sync: Boolean=None,
            json: Boolean=false,
            **options: Any,
    ) -> RequestHandler:
        name1 = name or self.program.name
        sync1 = self.sync_default if sync is None else sync
        return RequestHandler(
            self.method,
            self.program,
            name1,
            prefix,
            sync1,
            json,
            Map(options),
        )


class RequestHandlers(ToStr):

    @staticmethod
    def cons(*handlers: RequestHandler) -> 'RequestHandlers':
        return RequestHandlers(Map(Lists.wrap(handlers).apzip(_.name).map2(flip)))

    def __init__(self, handlers: Map[str, RequestHandler]) -> None:
        self.handlers = handlers

    def _arg_desc(self) -> List[str]:
        return List(str(self.handlers))

    def rpc_specs(self, name: str, prefix: str) -> List[RpcHandlerSpec]:
        return self.handlers.v / __.spec(name, prefix)


class rpc:

    def __new__(self, prog: P) -> RequestHandlerBuilder[CmdMethod]:
        return RequestHandlerBuilder(CmdMethod(), prog)

    @staticmethod
    def simple(f: Callable[..., NvimIO[A]]) -> RequestHandlerBuilder[CmdMethod]:
        ...



__all__ = ('RequestHandler', 'RequestHandlers')



from typing import Callable, TypeVar, Generic

from amino import Dat, List, Nil

from ribosome.config.settings import Settings

D = TypeVar('D')
S = TypeVar('S', bound=Settings)


class NoData(Dat['NoData']):
    pass


class BasicConfig(Generic[S, D], Dat['BasicConfig[D]']):

    @staticmethod
    def cons(
            name: str,
            prefix: str=None,
            state_ctor: Callable[[], D]=None,
            settings: S=None,
            core_components: List[str]=Nil,
            default_components: List[str]=Nil,
            internal_component: bool=True,
    ) -> 'BasicConfig':
        return BasicConfig(
            name,
            prefix or name,
            state_ctor or NoData,
            settings or Settings(name=name),
            core_components.cons('internal') if internal_component else core_components,
            default_components,
        )

    def __init__(
            self,
            name: str,
            prefix: str,
            state_ctor: Callable,
            settings: S,
            core_components: List[str],
            default_components: List[str],
    ) -> None:
        self.name = name
        self.prefix = prefix
        self.state_ctor = state_ctor
        self.settings = settings
        self.core_components = core_components
        self.default_components = default_components


__all__ = ('NoData', 'BasicConfig')



from ribosome.test.integration.klk import VimIntegrationKlkSpec
from ribosome.test.config import default_config_name


class ExternalSpec(VimIntegrationKlkSpec):

    def plugin_name(self) -> str:
        return default_config_name


__all__ = ('ExternalSpec',)



from kallikrein import Expectation

from ribosome.test.integration.klk import AutoPluginIntegrationKlkSpec
from ribosome.config.config import NoData
from ribosome.config.settings import Settings
from ribosome.nvim.api.variable import variable_set_prefixed
from ribosome.nvim.api.command import doautocmd

from integration._support.autocmd import val


class AutocmdSpec(AutoPluginIntegrationKlkSpec[Settings, NoData]):
    '''
    execute handler when triggering an autocmd $autocmd
    '''

    def plugin_name(self) -> str:
        return 'plug'

    def module(self) -> str:
        return 'integration._support.autocmd'

    def _pre_start(self) -> None:
        super()._pre_start()
        variable_set_prefixed('components', ['core']).unsafe(self.vim)

    def autocmd(self) -> Expectation:
        self._wait(2)
        doautocmd('VimResized').unsafe(self.vim)
        return self.var_becomes('autocmd_success', val)


__all__ = ('AutocmdSpec',)



from typing import Generic, TypeVar, Type

from lenses import UnboundLens

from amino import Dat

from ribosome.config.settings import Settings
from ribosome.data.plugin_state import PluginState

S = TypeVar('S', bound=Settings)
D = TypeVar('D')
CC = TypeVar('CC')
C = TypeVar('C')


class Ribosome(Generic[S, D, CC, C], Dat['Ribosome[S, D, CC, C]']):

    def __init__(
            self,
            state: PluginState[S, D, CC],
            comp_type: Type[C],
            comp_lens: UnboundLens['Ribosome[S, D, CC, C]', 'Ribosome[S, D, CC, C]', C, C],
    ) -> None:
        self.state = state
        self.comp_type = comp_type
        self.comp_lens = comp_lens


__all__ = ('Ribosome',)



from typing import TypeVar, Generic

from amino.dat import Dat

from ribosome.config.settings import Settings
from ribosome.config.component import Components

A = TypeVar('A')
D = TypeVar('D')
S = TypeVar('S', bound=Settings)
CC = TypeVar('CC')
G = TypeVar('G')


class Resources(Generic[S, D, CC], Dat['Resources[S, D, CC]']):

    def __init__(self, data: D, settings: S, components: Components[D, CC]) -> None:
        self.data = data
        self.settings = settings
        self.components = components


__all__ = ('Resources',)



from typing import TypeVar, Callable, Any

from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.data import NResult, NSuccess

from kallikrein import kf
from kallikrein.expectable import Expectable
from kallikrein.matchers.contain import PredContain, NestContain
from kallikrein.match_result import MatchResult, ContainsMatchResult, SimpleMatchResult
from kallikrein.matcher import BoundMatcher

from amino import Boolean, List
from amino.boolean import false

A = TypeVar('A')


def kn(vim: NvimApi, f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> Expectable:
    return kf(lambda: f(*a, **kw).result(vim))


def kns(vim: NvimApi, f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> Expectable:
    return kf(lambda: f(*a, **kw).run_s(vim))


class PredContainNResult(PredContain, tpe=NResult):

    def check(self, exp: NResult[A], target: A) -> Boolean:
        return Boolean(isinstance(exp, NSuccess) and exp.value == target)


class NestContainNResult(NestContain, tpe=NResult):

    def match(self, exp: NResult[A], target: BoundMatcher[A]) -> MatchResult[A]:
        return (
            target.evaluate(exp.value)
            if isinstance(exp, NSuccess) else
            SimpleMatchResult(false, List('{exp} is not NSuccess'))
        )

    def wrap(self, name: str, exp: NResult[A], nested: MatchResult[A]) -> MatchResult[A]:
        return ContainsMatchResult(name, exp, nested)


__all__ = ('kn',)



from typing import Callable, TypeVar

from ribosome.config.settings import Settings

from amino import Either
from ribosome.config.setting import Setting
from ribosome.nvim.io.state import NS
from ribosome.config.resources import Resources
from ribosome.data.plugin_state import PluginState
from ribosome.nvim.io.compute import NvimIO

A = TypeVar('A')
S = TypeVar('S', bound=Settings)
D = TypeVar('D')
CC = TypeVar('CC')


def setting(attr: Callable[[S], Setting[A]]) -> NS[Resources[S, D, CC], A]:
    def get(res: Resources[S, D, CC]) -> NvimIO[A]:
        return attr(res.settings).value_or_default
    return NS.inspect_f(get)


def setting_ps(attr: Callable[[S], Setting[A]]) -> NS[PluginState[S, D, CC], A]:
    def get(ps: PluginState[S, D, CC]) -> NvimIO[A]:
        return attr(ps.basic.settings).value_or_default
    return NS.inspect_f(get)


def setting_ps_e(attr: Callable[[S], Setting[A]]) -> NS[PluginState[S, D, CC], Either[str, A]]:
    def get(ps: PluginState[S, D, CC]) -> NvimIO[Either[str, A]]:
        return attr(ps.basic.settings).value
    return NS.inspect_f(get)


__all__ = ('setting', 'setting_ps', 'setting_ps_e')



from typing import TypeVar, Any, Generic

from amino import List, Boolean
from amino.dat import Dat

from ribosome.data.plugin_state import Programs
from ribosome.data.plugin_state_holder import PluginStateHolder

D = TypeVar('D')


class RequestJob(Generic[D], Dat['RequestJob']):

    def __init__(
            self,
            state: PluginStateHolder[D],
            name: str,
            args: List[Any],
            sync: bool,
            bang: Boolean,
    ) -> None:
        self.state = state
        self.name = name
        self.args = args
        self.sync = sync
        self.bang = Boolean(bang)

    @property
    def plugin_name(self) -> str:
        return self.state.state.basic.name

    @property
    def sync_prefix(self) -> str:
        return '' if self.sync else 'a'

    @property
    def desc(self) -> str:
        return f'{self.sync_prefix}sync request {self.name}({self.args}) to `{self.plugin_name}`'

    @property
    def programs(self) -> Programs:
        return self.state.state.programs


__all__ = ('RequestJob',)



import abc
from typing import TypeVar, Callable, Any, Generic

from amino.tc.base import TypeClass, tc_prop
from amino import Either, __, IO, Maybe
from amino.state import tcs, StateT, State, EitherState
from amino.func import CallByName

from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N

A = TypeVar('A')
S = TypeVar('S')


class NvimIOState(Generic[S, A], StateT[NvimIO, S, A], tpe=NvimIO):

    @staticmethod
    def io(f: Callable[[NvimApi], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.delay(f))

    @staticmethod
    def delay(f: Callable[[NvimApi], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.delay(f))

    @staticmethod
    def suspend(f: Callable[[NvimApi], NvimIO[A]]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.suspend(f))

    @staticmethod
    def from_io(io: IO[A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.wrap_either(lambda v: io.attempt))

    @staticmethod
    def from_id(st: State[S, A]) -> 'NvimIOState[S, A]':
        return st.transform_f(NvimIOState, lambda s: N.pure(s.value))

    @staticmethod
    def from_maybe(a: Maybe[A], err: CallByName) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.from_maybe(a, err))

    m = from_maybe

    @staticmethod
    def from_either(e: Either[str, A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.from_either(e))

    e = from_either

    @staticmethod
    def from_either_state(st: EitherState[S, A]) -> 'NvimIOState[S, A]':
        return st.transform_f(NvimIOState, lambda s: N.from_either(s))

    @staticmethod
    def failed(e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.failed(e))

    @staticmethod
    def error(e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(N.error(e))

    @staticmethod
    def inspect_maybe(f: Callable[[S], Either[str, A]], err: CallByName) -> 'NvimIOState[S, A]':
        return NvimIOState.inspect_f(lambda s: N.from_maybe(f(s), err))

    @staticmethod
    def inspect_either(f: Callable[[S], Either[str, A]]) -> 'NvimIOState[S, A]':
        return NvimIOState.inspect_f(lambda s: N.from_either(f(s)))

    @staticmethod
    def call(name: str, *args: Any, **kw: Any) -> 'NvimIOState[S, A]':
        return NS.delay(__.call(name, *args, **kw))


tcs(NvimIO, NvimIOState)

NS = NvimIOState


class ToNvimIOState(TypeClass):

    @abc.abstractproperty
    def nvim(self) -> NS:
        ...


class IdStateToNvimIOState(ToNvimIOState, tpe=State):

    @tc_prop
    def nvim(self, fa: State[S, A]) -> NS:
        return NvimIOState.from_id(fa)


class EitherStateToNvimIOState(ToNvimIOState, tpe=EitherState):

    @tc_prop
    def nvim(self, fa: EitherState[S, A]) -> NS:
        return NvimIOState.from_either_state(fa)


__all__ = ('NvimIOState', 'NS')



from typing import TypeVar, Callable, Any

from amino import Either, Boolean, do, Do
from amino.state import State

from ribosome.nvim.api.data import NvimApi
from ribosome.nvim.io.compute import NvimIOSuspend, NvimIOPure, NvimIOError, NvimIO, NvimIORecover, lift_n_result
from ribosome.nvim.io.data import NError, NFatal, NResult

A = TypeVar('A')
B = TypeVar('B')


def nvimio_delay(f: Callable[..., A], *a: Any, **kw: Any) -> NvimIO[A]:
    return NvimIOSuspend.cons(State.inspect(lambda vim: NvimIOPure(f(vim, *a, **kw))))


def nvimio_suspend(f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> NvimIO[A]:
    return NvimIOSuspend.cons(State.inspect(lambda vim: f(vim, *a, **kw)))


def nvimio_recover_error(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
    return NvimIORecover(fa, f, Boolean.is_a(NError))


def nvimio_recover_fatal(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
    return NvimIORecover(fa, f, Boolean.is_a(NFatal))


def nvimio_recover_failure(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[A]]) -> NvimIO[A]:
    return NvimIORecover(fa, f, Boolean.is_a((NError, NFatal)))


def nvimio_ensure(fa: NvimIO[A], f: Callable[[NResult[A]], NvimIO[None]], pred: Callable[[NResult[A]], Boolean]
                  ) -> NvimIO[A]:
    @do(NvimIO[A])
    def effect(error: NResult[A]) -> Do:
        yield f(error)
        yield lift_n_result.match(error)
    return NvimIORecover(fa, effect, pred)


def nvimio_wrap_either(f: Callable[[NvimApi], Either[B, A]]) -> NvimIO[A]:
    return nvimio_suspend(lambda v: f(v).cata(NvimIOError, NvimIOPure))


def nvimio_from_either(e: Either[str, A]) -> NvimIO[A]:
    return e.cata(NvimIOError, NvimIOPure)


__all__ = ('nvimio_delay', 'nvimio_wrap_either', 'nvimio_from_either', 'nvimio_suspend', 'nvimio_recover_error',
           'nvimio_recover_fatal', 'nvimio_recover_failure', 'nvimio_ensure')



import abc
from uuid import uuid4, UUID

from amino import Dat, Map, Boolean, ADT, List, Maybe
from amino.boolean import false

from ribosome.compute.program import Program


class MapMode(ADT['MapMode']):

    @abc.abstractproperty
    def mnemonic(self) -> str:
        ...


class mapmode:

    class Normal(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'n'

    class Operator(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'o'

    class Visual(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'x'

    class Select(MapMode):

        @property
        def mnemonic(self) -> str:
            return 's'

    class VisualSelect(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'v'

    class Insert(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'i'

    class CommandLine(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'c'

    class Terminal(MapMode):

        @property
        def mnemonic(self) -> str:
            return 't'

    class Language(MapMode):

        @property
        def mnemonic(self) -> str:
            return 'l'


class Mapping(Dat['Mapping']):

    @staticmethod
    def cons(keys: str, buffer: Boolean=false, modes: List[MapMode]=None, uuid: UUID=None) -> 'Mapping':
        modes1 = List(mapmode.Normal()) if modes is None else modes
        return Mapping(keys, Boolean(buffer), modes1, uuid or uuid4())

    def __init__(self, keys: str, buffer: Boolean, modes: List[MapMode], uuid: UUID) -> None:
        self.keys = keys
        self.buffer = buffer
        self.modes = modes
        self.uuid = uuid


class Mappings(Dat['Mappings']):

    @staticmethod
    def cons(mappings: Map[Mapping, Program]=Map()) -> 'Mappings':
        return Mappings(mappings)

    def __init__(self, mappings: Map[Mapping, Program]) -> None:
        self.mappings = mappings

    def lift(self, mapping: Mapping) -> Maybe[Program]:
        return self.mappings.lift(mapping)


__all__ = ('Mapping', 'Mappings')



from typing import TypeVar, Generic, Callable, Any

from amino import ADT, Dat, List, Lists, Nil
from amino.case import Case

from ribosome.compute.wrap_data import ProgWrappers
from ribosome.compute.output import ProgOutput, ProgOutputResult
from ribosome.nvim.io.state import NS
from ribosome.request.args import ParamsSpec
from ribosome.compute.prog import Prog, ProgBind, ProgExec

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
R = TypeVar('R')


class ProgramCode(Generic[A], ADT['ProgramCode[A]']):
    pass


# TODO restrict the function to a single value parameter of type ProgramArgs[A, B], where A is a custom data type
# containing the positional args and B an optional json decoded parameter.
class ProgramBlock(Generic[A, B, D], ProgramCode[B]):

    def __init__(
            self,
            code: Callable[..., NS[D, A]],
            wrappers: ProgWrappers,
            interpreter: ProgOutput[A, B],
    ) -> None:
        self.code = code
        self.wrappers = wrappers
        self.interpreter = interpreter


class ProgramCompose(Generic[A], ProgramCode[A]):

    def __init__(self, code: Callable[..., ProgBind[Any, A]]) -> None:
        self.code = code


class Program(Generic[A], Dat['Program[A]']):

    @staticmethod
    def lift(f: Callable[..., NS[D, A]]) -> 'Program[A]':
        return Program(
            f.__name__,
            ProgramBlock(f, ProgWrappers.id(), ProgOutputResult()),
            ParamsSpec.from_function(f)
        )

    def __init__(
            self,
            name: str,
            code: ProgramCode[A],
            params_spec: ParamsSpec,
    ) -> None:
        self.name = name
        self.code = code
        self.params_spec = params_spec

    def __call__(self, *args: Any) -> Prog[A]:
        return bind_program(self, Lists.wrap(args))


class bind_program_code(Generic[A, B, R], Case[ProgramCode[A], Prog[B]], alg=ProgramCode):

    def __init__(self, program: Program[A], args: List[Any]) -> None:
        self.args = args

    def program_compose(self, code: ProgramCompose[A]) -> Prog[A]:
        return code.code(*self.args)

    def program_block(self, code: ProgramBlock[Any, A, R]) -> Prog[A]:
        return ProgExec(code.code(*self.args), code.wrappers, code.interpreter)


def bind_program(program: Program[A], args: List[Any]) -> Prog[B]:
    return bind_program_code(program, args)(program.code)


def bind_nullary_program(program: Program[A]) -> Prog[B]:
    return bind_program(program, Nil)


def bind_programs(programs: List[Program[A]], args: List[Any]) -> Prog[B]:
    return programs.traverse(lambda a: bind_program(a, args), Prog)


__all__ = ('Program', 'bind_program', 'bind_nullary_program', 'bind_programs')



from typing import TypeVar, Type, Callable, Any
from types import ModuleType

from neovim.msgpack_rpc import MsgpackStream, AsyncSession, Session
from neovim.api import Nvim
from neovim.msgpack_rpc.event_loop.base import BaseEventLoop
from neovim.msgpack_rpc.event_loop.uv import UvEventLoop

from amino import Either, _, L, amino_log, Logger, __, Path, Nil, Just, IO
from amino.either import ImportFailure
from amino.logging import amino_root_file_logging
from amino.do import do, Do
from amino.mod import instance_from_module

from ribosome.config.config import Config
from ribosome.request.execute import execute_request
from ribosome.logging import ribo_log, nvim_logging
from ribosome.nvim.api.data import NvimApi, NativeNvimApi
from ribosome.data.plugin_state import PluginState
from ribosome.config.settings import Settings
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.variable import variable_set_prefixed
from ribosome.nvim.io.api import N
from ribosome.data.plugin_state_holder import PluginStateHolder
from ribosome.nvim.io.state import NS
from ribosome.config.component import ComponentConfig
from ribosome.components.internal.update import init_rpc
from ribosome.compute.output import ProgOutput
from ribosome.compute.prog import Prog
from ribosome.compute.interpret import ProgIOInterpreter
from ribosome.compute.program import Program

Loop = TypeVar('Loop', bound=BaseEventLoop)
D = TypeVar('D')
DIO = TypeVar('DIO')
B = TypeVar('B')
C = TypeVar('C', bound=Config)
S = TypeVar('S', bound=Settings)
CC = TypeVar('CC')
R = TypeVar('R')


def request_handler(vim: NvimApi, sync: bool, state: PluginStateHolder[D]) -> Callable[[str, tuple], Any]:
    def handle(name: str, args: tuple) -> Any:
        try:
            return execute_request(vim, state, name, args, sync)
        except Exception as e:
            desc = f'dispatching request {name}({args})'
            amino_log.caught_exception(desc, e)
            ribo_log.error(f'fatal error {desc}')
    return handle


@do(NvimIO[PluginState[S, D, CC]])
def init_state(config: Config, io_interpreter: ProgIOInterpreter=None, logger: Program[None]=None) -> Do:
    data = config.basic.state_ctor()
    log_handler = yield N.delay(nvim_logging)
    state = PluginState.cons(
        config.basic,
        ComponentConfig(config.components),
        config.request_handlers,
        data,
        Nil,
        config.init,
        logger=logger,
        io_interpreter=io_interpreter,
        log_handler=log_handler,
    )
    yield init_rpc().run_s(state)


@do(NvimIO[int])
def run_session(session: Session, config: Config) -> Do:
    state = yield init_state(config)
    holder = PluginStateHolder.concurrent(state)
    ribo_log.debug(f'running session')
    yield N.from_io(IO.delay(session._enqueue_notification, 'function:internal_init', ()))
    yield variable_set_prefixed('started', True)
    yield N.delay(
        lambda vim:
        session.run(
            request_handler(vim, True, holder),
            request_handler(vim, False, holder)
        )
    )
    yield N.pure(0)


def no_listen_address(err: Exception) -> None:
    raise Exception('could not connect to the vim server from within the host')


def run_loop(session: Session, config: Config) -> int:
    vim = NativeNvimApi(config.basic.name, Nvim.from_session(session)._session)
    return run_session(session, config).unsafe(vim)


def session(*args: str, loop: Type[Loop]=UvEventLoop, transport_type: str='stdio', **kwargs: str) -> Session:
    return Session(AsyncSession(MsgpackStream(loop(transport_type, *args, **kwargs))))


def config_from_module(mod: ModuleType) -> Either[str, Type[C]]:
    return instance_from_module(mod, Config)


log_initialized = False


def nvim_stdio_with_logging(name: str) -> NvimApi:
    from ribosome.logging import nvim_logging
    vim = NvimApi(name, session())
    nvim_logging(vim)
    return vim


def nvim_log() -> Logger:
    global log_initialized
    if not log_initialized:
        nvim_stdio_with_logging('ribosome_start_host')
        log_initialized = True
    return ribo_log


def config(config: Config) -> Config:
    return Config.cons(config.basic.name, config.state_ctor)


def start_config_stage_2(config: Config) -> int:
    amino_log.debug(f'starting plugin from {config.basic.name}')
    return run_loop(session(), config)


def error(msg: str) -> int:
    try:
        amino_log.error(msg)
        nvim_log().error(msg)
    except Exception as e:
        pass
    return 1


def import_error(e: ImportFailure, desc: str) -> int:
    return error(e.expand.join_lines)


def exception(e: Exception, desc: str) -> int:
    f = __.caught_exception_error(f'starting host from {desc}', e)
    try:
        f(amino_log)
        f(nvim_log())
    except Exception as e:
        pass
    return 1


def start_config_stage_1(mod: ModuleType) -> int:
    return config_from_module(mod).cata(error, start_config_stage_2)


def setup_log() -> None:
    amino_root_file_logging()


def start_from(source: str, importer: Callable[[str], Either[ImportFailure, ModuleType]], desc: str) -> int:
    try:
        setup_log()
        amino_log.debug(f'start_{desc}: {source}')
        return importer(source).cata(L(import_error)(_, source), start_config_stage_1)
    except Exception as e:
        return exception(e, source)


def start_module(mod: str) -> int:
    return start_from(mod, Either.import_module, 'module')


def start_file(path: str) -> int:
    p = Path(path)
    file = p / '__init__.py' if p.is_dir() else p
    return start_from(str(file), Either.import_file, 'file')


__all__ = ('start_host', 'start_module', 'start_file')



from typing import TypeVar, Generic, Any
import logging

from amino import ADT, IO, List, Dat, Nil
from amino.dat import DatMeta
from ribosome.process import Subprocess

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
P = TypeVar('P')
PIO = TypeVar('PIO')


class ProgOutput(Generic[A, B], ADT['ProgOutput[A, B]']):
    pass


class ProgOutputUnit(ProgOutput[None, None]):
    pass


class ProgOutputResult(Generic[A], ProgOutput[A, A]):
    pass


class ProgIO(Generic[PIO], ADT['ProgIO[PIO]']):
    pass


class ProgScalarIO(ProgIO[IO]):
    pass


class ProgGatherIOs(ProgIO[List[IO]]):
    pass


class ProgScalarSubprocess(ProgIO[Subprocess]):
    pass


class ProgGatherSubprocesses(ProgIO[List[Subprocess]]):
    pass


class EchoMeta(DatMeta):

    @property
    def unit(self) -> 'Echo':
        return Echo(Nil, 0)


class Echo(Dat['Echo'], metaclass=EchoMeta):
    info_level = logging.INFO
    error_level = logging.ERROR

    @staticmethod
    def info(msg: str) -> 'Echo':
        return Echo(List(msg), Echo.info_level)

    @staticmethod
    def error(msg: str) -> 'Echo':
        return Echo(List(msg), Echo.error_level)

    def __init__(self, messages: List[str], level: int) -> None:
        self.messages = messages
        self.level = level


class ProgIOEcho(ProgIO[None]):
    pass


class ProgIOCustom(ProgIO[Any]):
    pass


class ProgOutputIO(Generic[PIO, A], ProgOutput[PIO, A]):

    def __init__(self, io: ProgIO[PIO]) -> None:
        self.io = io


class GatherIOs(Generic[A], Dat['GatherIOs[A]']):

    def __init__(self, ios: List[IO[A]], timeout: float) -> None:
        self.ios = ios
        self.timeout = timeout


class GatherSubprocesses(Generic[A], Dat['Generic[A]']):

    def __init__(self, subprocs: List[Subprocess[A]], timeout: float) -> None:
        self.subprocs = subprocs
        self.timeout = timeout


__all__ = ('ProgOutput', 'ProgOutputUnit', 'ProgOutputResult', 'ProgOutputIO', 'ProgResult', 'ProgReturn',
           'ProgIO', 'ProgScalarIO', 'ProgGatherIOs', 'ProgScalarSubprocess', 'ProgGatherSubprocesses',
           'ProgIOCustom', 'ProgOutputIO', 'GatherIOs', 'GatherSubprocesses')



from kallikrein import Expectation

from ribosome.test.integration.klk import AutoPluginIntegrationKlkSpec
from ribosome.config.config import NoData
from ribosome.config.settings import Settings
from ribosome.nvim.api.variable import variable_set
from ribosome.nvim.api.exists import command_once_defined


class SettingsSpec(AutoPluginIntegrationKlkSpec[Settings, NoData]):
    '''
    update a setting $update
    '''

    def plugin_name(self) -> str:
        return 'plug'

    def module(self) -> str:
        return 'integration._support.settings'

    def _pre_start(self) -> None:
        super()._pre_start()
        variable_set('counter', 7).unsafe(self.vim)
        variable_set('inc', 14).unsafe(self.vim)

    def update(self) -> Expectation:
        command_once_defined('PlugCheck').unsafe(self.vim)
        self._wait(.5)
        return self.var_becomes('counter', 21)


__all__ = ('SettingsSpec',)



from amino import List, do, Do, _, __

from ribosome.compute.api import prog
from ribosome.request.handler.handler import RequestHandler
from ribosome.config.config import Config, NoData
from ribosome.nvim.io.state import NS
from ribosome.config.settings import Settings, int_setting
from ribosome.config.component import NoComponentData
from ribosome.compute.ribosome import Ribosome
from ribosome.compute.ribosome_api import Ribo


class PlugSettings(Settings):

    def __init__(self) -> None:
        super().__init__('plug')
        self.counter = int_setting('counter', 'counter', '', False)
        self.inc = int_setting('inc', 'inc', '', False)


@prog.unit
@do(NS[Ribosome[Settings, NoData, NoComponentData, NoData], None])
def check() -> Do:
    counter = yield Ribo.setting(_.counter)
    inc = yield Ribo.setting(_.inc)
    yield NS.inspect_f(__.state.basic.settings.counter.update(counter + inc))


config: Config = Config.cons(
    name='plug',
    request_handlers=List(
        RequestHandler.trans_cmd(check)(),
    ),
    settings=PlugSettings()
)

__all__ = ('config',)



from typing import Any, TypeVar

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.util import cons_split_lines
from ribosome.nvim.io.api import N

A = TypeVar('A')


def nvim_command(cmd: str, *args: Any, verbose: bool=False) -> NvimIO[None]:
    arg_string = ' '.join(map(str, args))
    arg_suffix = '' if len(args) == 0 else f' {arg_string}'
    silent = '' if verbose else 'silent! '
    return N.write('nvim_command', f'{silent}{cmd}{arg_suffix}')


def nvim_command_output(cmd: str, *args: Any) -> NvimIO[None]:
    arg_string = ' '.join(map(str, args))
    return N.read_cons('nvim_command_output', cons_split_lines, f'{cmd} {arg_string}')


def doautocmd(name: str, pattern: str='', verbose: bool=False) -> NvimIO[None]:
    return nvim_command(f'doautocmd', '<nomodeline>', name, pattern, verbose=verbose)


def runtime(path: str, verbose: bool=True) -> NvimIO[None]:
    return nvim_command('runtime!', f'{path}.vim', verbose=verbose)


__all__ = ('nvim_command', 'nvim_command_output', 'doautocmd', 'runtime')



from kallikrein import k, Expectation
from kallikrein.matchers import contain
from kallikrein.matchers.typed import have_type

from amino import do, Do

from ribosome.test.integration.default import ExternalSpec
from ribosome.test.klk import kn
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.rpc import channel_id
from ribosome.nvim.api.variable import variable_set_prefixed, variable_prefixed_num


class ApiSpec(ExternalSpec):
    '''
    channel id $channel_id
    set and get a prefixed variable $set_and_get_var
    '''

    def channel_id(self) -> Expectation:
        return kn(self.vim, channel_id).must(contain(have_type(int)))

    def set_and_get_var(self) -> Expectation:
        name = 'varname'
        value = 5
        @do(NvimIO[int])
        def go() -> Do:
            yield variable_set_prefixed(name, value)
            yield variable_prefixed_num(name)
        return kn(self.vim, go).must(contain(value))


__all__ = ('ApiSpec',)



from typing import TypeVar

from amino import do, Do, _, __, List, Either, L
from amino.lenses.lens import lens
from amino.state import EitherState

from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PluginState, Programs, handler_spec
from ribosome.config.resolve import ComponentResolver
from ribosome.request.handler.handler import RequestHandler, RequestHandlers
from ribosome.config.component import Components, Component
from ribosome.request.rpc import define_handlers, RpcHandlerSpec
from ribosome.config.settings import Settings
from ribosome.compute.program import Program
from ribosome.request.handler.method import RpcMethod
from ribosome.util.setting import setting_ps_e

CC = TypeVar('CC')
D = TypeVar('D')
Meth = TypeVar('Meth', bound=RpcMethod)
P = TypeVar('P')
S = TypeVar('S', bound=Settings)


def request_handlers_programs(handlers: RequestHandlers) -> List[Program]:
    return handlers.handlers.v


def component_programs(component: Component) -> List[Program]:
    return request_handlers_programs(component.request_handlers)


def rpc_method(name: str, prefix: str) -> str:
    def rpc_method(handler: RequestHandler) -> str:
        return handler_spec(handler, name, prefix).rpc_method
    return rpc_method


def programs(state: PluginState[S, D, CC]) -> Programs:
    cfg_programs = request_handlers_programs(state.request_handlers)
    compo_programs = state.components.all // component_programs
    programs = compo_programs + cfg_programs
    method = rpc_method(state.basic.name, state.basic.prefix)
    return programs.group_by(method)


@do(EitherState[PluginState[S, D, CC], None])
def update_components(from_user: Either[str, List[str]]) -> Do:
    name = yield EitherState.inspect(_.basic.name)
    components_map = yield EitherState.inspect(_.comp.available)
    core_components = yield EitherState.inspect(_.basic.core_components)
    default_components = yield EitherState.inspect(_.basic.default_components)
    resolver = ComponentResolver(name, components_map, core_components, default_components, from_user)
    components = yield EitherState.lift(resolver.run())
    yield EitherState.modify(__.copy(components=Components.cons(components)))
    progs = yield EitherState.inspect(programs)
    yield EitherState.modify(lens.programs.set(progs))


# TODO finish
@do(NS[PluginState[S, D, CC], None])
def undef_handler(handler: RpcHandlerSpec) -> Do:
    yield NS.unit


@do(NS[PluginState[S, D, CC], None])
def undef_handlers() -> Do:
    handlers = yield NS.inspect(_.rpc_handlers)
    yield handlers.traverse(undef_handler, NS)


@do(NS[PluginState[S, D, CC], None])
def def_handlers() -> Do:
    name = yield NS.inspect(_.basic.name)
    specs = yield NS.inspect(_.distinct_specs)
    handlers = yield NS.lift(define_handlers(specs, name))
    yield NS.modify(lens.rpc_handlers.set(handlers))


@do(NS[PluginState[S, D, CC], None])
def init_rpc() -> Do:
    from_user = yield setting_ps_e(_.components)
    yield update_components(from_user).nvim
    yield def_handlers()


__all__ = ('init_rpc', 'undef_handlers', 'def_handlers', 'programs')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x6a61a38b

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

from typing import TypeVar  # line 1

from lenses import UnboundLens  # line 3

from amino import IO  # line 5
from amino import List  # line 5
from amino import Path  # line 5
from amino import Left  # line 5
from amino import do  # line 5
from amino import Do  # line 5
from amino.json import decode_json  # line 6
from amino.json import dump_json  # line 6

from ribosome.nvim.io.compute import NvimIO  # line 8
from ribosome.nvim.io.state import NS  # line 9
from ribosome.config.resources import Resources  # line 10
from ribosome.config.settings import Settings  # line 11
from ribosome.nvim.io.api import N  # line 12

A = TypeVar('A')  # line 14
D = TypeVar('D')  # line 15
S = TypeVar('S', bound=Settings)  # line 16
CC = TypeVar('CC')  # line 17


settings: NS[Resources[S, D, CC], S] = NS.inspect(lambda a: a.settings)  # line 20


def mkdir(dir: 'Path') -> 'IO[None]':  # line 23
    return IO.delay(dir.mkdir, parents=True, exist_ok=True)  # line 24


@do(NvimIO[Path])  # line 27
def state_file(settings: 'S', name: 'str') -> 'Do':  # line 28
    dir = yield settings.project_state_dir.value_or_default  # line 29
    yield N.from_io(mkdir(dir))  # line 30
    yield N.pure(dir / f'{name}.json')  # line 31


@do(NvimIO[A])  # line 34
def load_json_data_from(name: 'str', file: 'Path') -> 'Do':  # line 35
    exists = yield N.from_io(IO.delay(file.exists))  # line 36
    if exists:  # line 37
        json = yield N.from_io(IO.delay(file.read_text))  # line 38
        yield N.pure(decode_json(json))  # line 39
    else:  # line 40
        yield N.pure(Left(f'state file {file} does not exist'))  # line 41


@do(NvimIO[A])  # line 44
def load_json_data(settings: 'S', name: 'str') -> 'Do':  # line 45
    file = yield state_file(settings, name)  # line 46
    yield load_json_data_from(file)  # line 47


@do(NS[Resources[S, D, CC], None])  # line 50
def load_json_state(name: 'str', store: 'UnboundLens') -> 'Do':  # line 51
    s = yield settings  # line 52
    state = yield NS.lift(load_json_data(s, name))  # line 53
    yield state.cata(lambda a: NS.pure(None), lambda d: NS.modify(store.set(d)))  # line 54


@do(NvimIO[None])  # line 57
def store_json_data(settings: 'S', name: 'str', data: 'A') -> 'Do':  # line 58
    file = yield state_file(settings, name)  # line 59
    json = yield N.from_either(dump_json(data))  # line 60
    yield N.from_io(IO.delay(file.write_text, json))  # line 61
    yield N.pure(None)  # line 62


@do(NS[Resources[S, D, CC], None])  # line 65
def store_json_state(name: 'str', fetch: '_coconut.typing.Callable[[D], A]') -> 'Do':  # line 66
    payload = yield NS.inspect(lambda s: fetch(s.data))  # line 67
    s = yield settings  # line 68
    yield NS.lift(store_json_data(s, name, payload))  # line 69

__all__ = ('load_json_state', 'store_json_data', 'store_json_state')  # line 71



from typing import GenericMeta, TypeVar, Callable, Type

from amino.lenses.lens import lens
from amino import do, Do, _

from ribosome.nvim.io.state import NS
from ribosome.config.settings import Settings
from ribosome.compute.ribosome import Ribosome
from ribosome.config.setting import Setting
from ribosome.nvim.io.compute import NvimIO
from ribosome.compute.api import prog
from ribosome.compute.prog import Prog, ProgExec
from ribosome.compute.output import ProgOutputResult
from ribosome.compute.tpe_data import StateProg, trivial_state_prog, ribo_state_prog
from ribosome.compute.wrap import prog_wrappers
from ribosome.data.plugin_state import PluginState

A = TypeVar('A')
S = TypeVar('S', bound=Settings)
D = TypeVar('D')
M = TypeVar('M')
CC = TypeVar('CC')
C = TypeVar('C')
R = TypeVar('R')
main_lens = lens.state.data


class RMeta(GenericMeta):
    pass


class Ribo(metaclass=RMeta):

    @classmethod
    def setting(self, attr: Callable[[S], Setting[A]]) -> NS[Ribosome[S, D, CC, C], A]:
        def get(rib: Ribosome[S, D, CC, C]) -> NvimIO[A]:
            return attr(rib.state.basic.settings).value_or_default
        return NS.inspect_f(get)

    @classmethod
    @prog
    def setting_prog(self, attr: Callable[[S], Setting[A]]) -> NS[Ribosome[S, D, CC, C], A]:
        def get(rib: Ribosome[S, D, CC, C]) -> NvimIO[A]:
            return attr(rib.state.basic.settings).value_or_default
        return NS.inspect_f(get)

    @classmethod
    def settings(self) -> NS[Ribosome[S, D, CC, C], S]:
        return NS.inspect(lambda a: a.ps.settings)

    @classmethod
    def comp(self) -> NS[Ribosome[S, D, CC, C], C]:
        return NS.inspect(lambda a: a.comp_lens.get()(a))

    @classmethod
    def inspect_comp(self, f: Callable[[C], A]) -> NS[Ribosome[S, D, CC, C], A]:
        return NS.inspect(lambda a: f(a.comp_lens.get()(a)))

    @classmethod
    def inspect_comp_e(self, f: Callable[[C], A]) -> NS[Ribosome[S, D, CC, C], A]:
        return NS.inspect_either(lambda a: f(a.comp_lens.get()(a)))

    @classmethod
    def modify_comp(self, f: Callable[[C], C]) -> NS[Ribosome[S, D, CC, C], None]:
        return NS.modify(lambda a: a.comp_lens.modify(f)(a))

    @classmethod
    def main(self) -> NS[Ribosome[S, D, CC, C], C]:
        return NS.inspect(lambda a: main_lens.get()(a))

    @classmethod
    def modify_main(self, f: Callable[[D], D]) -> NS[Ribosome[S, D, CC, C], None]:
        return NS.modify(lambda a: main_lens.modify(f)(a))

    @classmethod
    @do(NS[Ribosome[S, D, CC, C], A])
    def zoom_comp(self, fa: NS[C, A]) -> Do:
        lens = yield NS.inspect(_.comp_lens)
        yield fa.zoom(lens)

    @classmethod
    @do(Prog[A])
    def lift_state_prog(self, fa: NS[Ribosome[S, D, CC, C], A], state_type: StateProg[M, C, R, A]) -> Do:
        wrappers = yield Prog.from_either(prog_wrappers.match(state_type))
        yield ProgExec(fa, wrappers, ProgOutputResult())

    @classmethod
    @do(Prog[A])
    def lift(self, fa: NS[Ribosome[S, D, CC, C], A], comp: Type[C]) -> Do:
        state_type: StateProg[PluginState[S, D, CC], C, Ribosome[S, D, CC, C]] = ribo_state_prog(comp)
        yield Ribo.lift_state_prog(fa, state_type)

    @classmethod
    def lift_comp(self, fa: NS[C, A], comp: Type[C]) -> Prog[A]:
        return Ribo.lift(Ribo.zoom_comp(fa), comp)

    @classmethod
    def trivial(self, fa: NS[D, A]) -> Prog[A]:
        return Ribo.lift_state_prog(fa, trivial_state_prog)

    @classmethod
    def lift_nvimio(self, fa: NvimIO[A]) -> Prog[A]:
        return Ribo.trivial(NS.lift(fa))


__all__ = ('Ribosome', 'Ribo')



import time
from typing import Callable, TypeVar, Type, Any, List as TList, Tuple

from msgpack import ExtType

from amino import Either, Left, Right, List, Lists, Try, do, Maybe, Do, Just, Nothing

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N

A = TypeVar('A')
B = TypeVar('B')


def cons_checked_e(tpe: Type[A], cons: Callable[[A], Either[str, B]]) -> Callable[[Any], Either[str, B]]:
    def cons_data(data: Any) -> Either[str, A]:
        return cons(data) if isinstance(data, tpe) else Left(f'invalid nvim data for `{tpe}`: {data}')
    return cons_data


def cons_checked(tpe: Type[A], cons: Callable[[A], B]) -> Callable[[Any], Either[str, B]]:
    return cons_checked_e(tpe, lambda a: Right(cons(a)))


def cons_ext(cons: Callable[[ExtType], A]) -> Callable[[Any], Either[str, A]]:
    return cons_checked(ExtType, cons)


def cons_checked_list(tpe: Type[A], cons: Callable[[A], B]) -> Callable[[Any], Either[str, List[B]]]:
    def cons_data(data: Any) -> Either[str, A]:
        return (
            Lists.wrap(data).traverse(cons_checked(tpe, cons), Either)
            if isinstance(data, TList) else
            Left(f'invalid nvim data for `List[{tpe}]`: {data}')
        )
    return cons_data


def cons_ext_list(cons: Callable[[ExtType], A]) -> Callable[[Any], Either[str, List[A]]]:
    return cons_checked_list(ExtType, cons)


cons_decode_str = cons_checked_e((bytes, str), (lambda a: Try(a.decode) if isinstance(a, bytes) else Right(a)))


def check_str_list(a: list) -> Either[str, List[str]]:
    return Lists.wrap(a).traverse(cons_decode_str, Either)


cons_decode_str_list = cons_checked_e(list, check_str_list)


cons_split_lines = cons_checked(str, Lists.lines)


def check_bool(a: Any) -> Either[str, bool]:
    return Right(a == 1) if a == 1 or a == 0 else Left(f'nvim result is not a boolean: {a}')


cons_decode_bool = cons_checked_e(int, check_bool)


def extract_int_pair(data: list) -> Either[str, Tuple[int, int]]:
    data_l = Lists.wrap(data)
    @do(Maybe[Tuple[int, int]])
    def run() -> Do:
        first = yield data_l.lift(0)
        second = yield data_l.lift(1)
        yield Just((first, second)) if isinstance(first, int) and isinstance(second, int) else Nothing
    return run().to_either(lambda: f'invalid nvim data for int pair: {data}')


def split_option(value: str) -> List[str]:
    return Lists.split(value, ',')


cons_decode_str_list_option = cons_checked(str, split_option)


@do(NvimIO[A])
def nvimio_repeat_timeout(
        thunk: Callable[[], NvimIO[A]],
        check: Callable[[A], bool],
        error: str,
        timeout: float,
) -> Do:
    start = yield N.simple(time.time)
    @do(NvimIO[None])
    def recurse() -> Do:
        result = yield thunk()
        done = check(result)
        yield (
            N.pure(result)
            if done else
            N.error(error)
            if time.time() - start > timeout else
            recurse()
        )
    yield recurse()


__all__ = ('cons_checked_e', 'cons_checked', 'cons_ext', 'cons_checked_list', 'cons_ext_list', 'check_str_list',
           'cons_decode_str', 'cons_decode_str_list', 'extract_int_pair', 'split_option', 'cons_decode_str_list_option',
           'cons_split_lines', 'cons_decode_bool', 'nvimio_repeat_timeout')



from kallikrein import Expectation
from kallikrein.matchers import contain

from amino.test import temp_dir

from ribosome.test.integration.klk import AutoPluginIntegrationKlkSpec
from ribosome.config.settings import Settings
from ribosome.config.config import NoData
from ribosome.test.klk import kn
from ribosome.nvim.api.function import nvim_call_tpe
from ribosome.nvim.api.exists import call_once_defined

from integration._support.default_handler import class_name


class DefaultHandlerSpec(AutoPluginIntegrationKlkSpec[Settings, NoData]):
    '''
    append a directory to the plugin's `sys.path` $append_path
    '''

    def plugin_name(self) -> str:
        return 'plug'

    def module(self) -> str:
        return 'integration._support.default_handler'

    def append_path(self) -> Expectation:
        pkg = temp_dir('default_handler', 'pp', 'pkg')
        pp = pkg.parent
        file = pkg / '__init__.py'
        file.write_text(f'class {class_name}: pass')
        call_once_defined('PlugAppendPythonPath', str(pp)).unsafe(self.vim)
        return (
            kn(self.vim, nvim_call_tpe, str, 'TestPath').must(contain(class_name)) &
            kn(self.vim, nvim_call_tpe, str, 'PlugShowPythonPath').must(contain(contain(str(pp))))
        )


__all__ = ('DefaultHandlerSpec',)



from typing import Callable, Any, Optional
from contextlib import contextmanager

from amino import List, Map, Boolean, Maybe, Just, Nil
from amino.util.fun import format_funcall

# from ribosome.nvim import Buffer, Tab, Window
# from ribosome.nvim.api.data import NvimApi
# from ribosome.nvim.components import Options, Vars


# class MockOptions(Options):

#     def set(self, *a, **kw):
#         pass


# class MockVars(Vars):

#     def __init__(self, vim, vars: dict) -> None:
#         super().__init__(vim)
#         self.vars = vars
#         self.prefix = vim.prefix

#     def _get(self, name: str) -> Optional[Any]:
#         return self.vars.get(name)

#     def set(self, name, value):
#         self.vars[name] = value


# class MockNvimComponent(object):

#     def __init__(self, prefix, vars: dict) -> None:
#         self.prefix = prefix
#         self._vars = MockVars(self, vars)
#         self._options = MockOptions(self)

#     @property
#     def window(self):
#         return MockWindow(self.vim, None, self.prefix)

#     @property
#     def buffer(self):
#         return MockBuffer(self.vim, None, self.prefix)

#     @property
#     def tab(self):
#         return MockTab(self.vim, None, self.prefix)


# class MockTab(MockNvimComponent, Tab):
#     pass


# class MockWindow(MockNvimComponent, Window):
#     pass


# class MockBuffer(MockNvimComponent, Buffer):
#     pass


# class MockSession:

#     def request(self, *a, **kw) -> None:
#         pass


# class MockNvim:

#     def __init__(self, responses: Callable[[str], Any]) -> None:
#         self.responses = responses
#         self.channel_id = 1
#         self.request_log = Nil

#     @property
#     def types(self) -> Map[str, type]:
#         return Map()

#     def command(self, cmdline: str, **kw: Any) -> Any:
#         self.log_request(cmdline)
#         return self.responses(cmdline).get_or_fail(f'no response recorded for command {cmdline}')

#     def call(self, name: str, *a: Any, **kw: Any) -> Any:
#         rep = format_funcall(name, a, kw)
#         self.log_request(rep)
#         return self.responses(name).get_or_fail(f'no response recorded for call {rep}')

#     def async_call(self, f: Callable[['MockNvim'], None]) -> None:
#         pass

#     def log_request(self, line: str) -> None:
#         self.request_log = self.request_log.cat(line)

#     @property
#     def _session(self) -> MockSession:
#         return MockSession()


# class MockNvimFacade(MockNvimComponent, NvimApi):

#     def __init__(self, prefix: str='ribosome', vars: dict=dict(),
#                  responses: Callable[[str], Maybe[Any]]=lambda a: Just(0)) -> None:
#         MockNvimComponent.__init__(self, prefix, vars)
#         self.vim = MockNvim(responses)
#         self.target = self.vim

#     @property
#     def windows(self):
#         return List()

#     def switch_root(self, root):
#         pass

#     def async(self, f: Callable[['NvimApi'], Any], *a, **kw):
#         return f(self, *a, **kw)

#     @contextmanager
#     def main_event_loop(self):
#         yield None

#     def reload_windows(self):
#         pass

#     def function_exists(self, name: str) -> Boolean:
#         return self.vim.responses(name).present


__all__ = ('MockNvimFacade',)



from amino import Either, List, do, Do

from ribosome.config.config import Config, NoData
from ribosome.compute.api import prog
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.handler.prefix import Plain
from ribosome.nvim.io.state import NS

class_name = 'ZeeKlass'


@prog.result
def test_path() -> NS[NoData, str]:
    return NS.pure(Either.import_name('pkg', class_name).map(lambda a: a.__name__).value_or('failed'))


config: Config = Config.cons(
    'plug',
    request_handlers=List(
        RequestHandler.trans_function(test_path)(prefix=Plain(), sync=True),
    )
)

__all__ = ('config',)



import abc
from typing import Any, Callable, Tuple

from neovim.msgpack_rpc import Session

from msgpack import ExtType

from amino import List, Either, Map, Left, Try, Dat, Nil, do, Do
from amino.logging import module_log

log = module_log()


class NvimApi(Dat['NvimApi']):

    def __init__(self, name: str) -> None:
        self.name = name

    @abc.abstractmethod
    def request(self, method: str, args: List[Any]) -> Either[str, Tuple['NvimApi', Any]]:
        ...


class NativeNvimApi(NvimApi):

    def __init__(self, name: str, session: Session) -> None:
        self.name = name
        self.session = session

    @do(Either[str, Tuple['NvimApi', Any]])
    def request(self, method: str, args: List[Any]) -> Do:
        log.debug1(lambda: f'executing nvim request {method}({args.join_comma})')
        result = yield Try(self.session.request, method, *args)
        return self, result


StrictNvimHandler = Callable[['StrictNvimApi', str, List[Any]], Either[List[str], Tuple[NvimApi, Any]]]


def no_request_handler(vim: 'StrictNvimApi', method: str, args: List[Any]) -> Either[List[str], Tuple[NvimApi, Any]]:
    return Left(List('no request handler defined'))


class StrictNvimApi(NvimApi):

    @staticmethod
    def cons(
            name: str,
            vars: Map[str, Any]=Map(),
            request_handler: StrictNvimHandler=no_request_handler,
            request_log: List[Tuple[str, List[Any]]]=Nil,
    ) -> 'StrictNvimApi':
        return StrictNvimApi(name, vars, request_handler, request_log)

    def __init__(
            self,
            name: str,
            vars: Map[str, Any],
            request_handler: StrictNvimHandler,
            request_log: List[Tuple[str, List[Any]]],
    ) -> None:
        self.name = name
        self.vars = vars
        self.request_handler = request_handler
        self.request_log = request_log

    def request(self, method: str, args: List[Any]) -> Either[str, Tuple['NvimApi', Any]]:
        vim = self.append1.request_log((method, args))
        return self.request_handler(vim, method, args).accum_error_lift(self.try_var, vim, method, args)

    def try_var(self, vim: 'StrictNvimApi', method: str, args: List[Any]) -> Either[str, Any]:
        return (
            args
            .detach_head
            .to_either('no variable name given')
            .flat_map2(lambda h, t: vim.var(h))
            .map(lambda a: (vim, a))
            if method == 'nvim_get_var' else
            Left(f'not a variable request')
        )

    def var(self, name: str) -> Either[str, Any]:
        return self.vars.lift(name).to_either(f'no variable `{name}` defined')


class Tabpage(Dat['Tabpage']):

    def __init__(self, data: ExtType) -> None:
        self.data = data


class Window(Dat['Window']):

    def __init__(self, data: ExtType) -> None:
        self.data = data


class Buffer(Dat['Buffer']):
    def __init__(self, data: ExtType) -> None:
        self.data = data


__all__ = ('NvimApi', 'NativeNvimApi', 'StrictNvimApi', 'Tabpage', 'Window', 'StrictNvimHandler')



from typing import Tuple

from kallikrein import k, Expectation
from kallikrein.matchers.length import have_length

from amino import do, Do, List
from amino.boolean import true

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.test.integration.run import RequestHelper
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.scratch import CreateScratchBufferOptions, show_in_scratch_buffer
from ribosome.test.config import single_trans_config
from ribosome.test.integration.default import ExternalSpec
from ribosome.nvim.api.ui import buffer_content, buffers, windows, current_buffer
from ribosome.nvim.api.data import Window, Buffer
from ribosome.config.config import NoData


@prog.result
@do(NS[NoData, None])
def create_scratch(lines: List[str], options: CreateScratchBufferOptions) -> Do:
    yield NS.lift(show_in_scratch_buffer(lines, options))
    yield NS.unit


config = single_trans_config(create_scratch, json=true)


class ScratchSpec(ExternalSpec):
    '''
    launch a scratch buffer $launch
    '''

    def launch(self) -> Expectation:
        helper = RequestHelper.nvim(config, self.vim, 'main')
        lines = List('1', '2', '3')
        @do(NvimIO[Tuple[List[Window], List[Buffer], List[str]]])
        def run() -> Do:
            yield helper.run('command:create_scratch', args=(lines, '{}',))
            wins = yield windows()
            bufs = yield buffers()
            buf = yield current_buffer()
            content = yield buffer_content(buf)
            return wins, bufs, content
        wins, bufs, content = run().unsafe(self.vim)
        return k(wins).must(have_length(2)) & k(bufs).must(have_length(2)) & (k(content) == lines)


__all__ = ('ScratchSpec',)



import inspect
from typing import Callable, Iterable, cast, Generator, TypeVar

from amino.options import env_xdg_data_dir
from amino import List, Either, Lists, Map, _, L, __, Path, Right, Try, Nil, Boolean, Left, Eval
from amino.func import flip
from amino.util.string import ToStr
from amino.do import do
from amino.boolean import true

from ribosome.nvim.io.compute import NvimIO
from ribosome.logging import Logging
from ribosome.config.setting import Setting, EvalSetting, setting_ctor
from ribosome.nvim.io.api import N

A = TypeVar('A')
B = TypeVar('B')
state_dir_help_default = '''This directory is used to persist the plugin's current state.'''
state_dir_base_default = env_xdg_data_dir.value / Path | (Path.home() / '.local' / 'share')
proteome_name_help = 'If **proteome** is installed, the session name is obtained from the main project name.'
session_name_help = 'A custom session name for the state dir can be specified.'
components_help = '''The plugin can run an arbitrary set of sub-components that inherit the class `Component`.
They receive all messages the core of the plugin processes and have access to the plugin state.
They can define nvim request handlers with the decorator `trans.xyz`.
The entries of this setting can either be names of the builtin components or arbitrary python module paths that define
custom components.
'''


@do(Either[str, str])
def project_name_from_path() -> Generator:
    cwd = yield Try(Path.cwd)
    name = cwd.name
    yield Right(name) if name else Left('cwd broken')


@do(NvimIO[Either[str, Path]])
def state_dir_with_name(state_dir: Setting[Path], proteome_name: Setting[str],
                        session_name: Setting[str]) -> Generator:
    base = yield state_dir.value_or_default
    pro_name = yield proteome_name.value
    sess_name = yield session_name.value
    path = pro_name.o(sess_name).o(project_name_from_path) / (lambda a: base / a)
    yield N.pure(path)


class Settings(ToStr, Logging):

    def __init__(self, name: str, state_dir_help: str=state_dir_help_default) -> None:
        self.name = name
        self.components = list_setting('components', 'names or paths of active components', components_help, True,
                                       Right(Nil))
        self.state_dir = path_setting('state_dir', 'state persistence directory', state_dir_help, True,
                                      Right(state_dir_base_default / self.name))
        self.proteome_name = str_setting('proteome_main_name', 'project name from protoeome', proteome_name_help, False)
        self.session_name = str_setting('ribosome_session_name', 'project name from user var', session_name_help, False)
        self.project_state_dir = EvalSetting(
            'project_state_dir',
            Eval.always(lambda: state_dir_with_name(self.state_dir, self.proteome_name, self.session_name))
        )
        self.run_internal_init = bool_setting('run_internal_init', 'run internal initialization handler', '', True,
                                              Right(true))

    def all(self) -> Map[str, Setting]:
        settings = inspect.getmembers(self, L(isinstance)(_, Setting))
        return Map(Lists.wrap(settings).map(_[1]).apzip(_.name).map2(flip))

    def _arg_desc(self) -> List[str]:
        return List(str(self.all()))


def path_list(data: list) -> Either[str, List[Path]]:
    return Lists.wrap(data).traverse(lambda a: Try(Path, a) / __.expanduser(), Either)


str_setting = setting_ctor(str, Right)
int_setting = setting_ctor(int, Right)
float_setting = setting_ctor(float, Right)
list_setting = setting_ctor(list, Right)
path_setting = setting_ctor(str, (lambda a: Try(Path, a)))
path_list_setting = setting_ctor(list, path_list)
map_setting = setting_ctor(dict, lambda a: Right(Map(a)))
path_map_setting = setting_ctor(dict, lambda a: Try(Map, a).valmap(lambda b: Path(b).expanduser()))
bool_setting = setting_ctor(int, lambda a: Right(Boolean(a)))


__all__ = ('Settings', 'str_setting', 'float_setting', 'list_setting', 'path_setting', 'path_list_setting',
           'map_setting', 'path_map_setting', 'bool_setting')



from amino.boolean import true
from amino import List

from ribosome.request.handler.handler import RequestHandler
from ribosome.components.internal.prog import (program_log, set_log_level, update_state, update_component_state,
                                               state_data, rpc_handlers, poll, append_python_path, show_python_path,
                                               enable_components, mapping, internal_init)
from ribosome.request.handler.prefix import Full
from ribosome.config.component import Component


program_log_handler = RequestHandler.trans_function(program_log)(prefix=Full(), sync=true)
set_log_level_handler = RequestHandler.trans_function(set_log_level)(prefix=Full())
update_state_handler = RequestHandler.trans_cmd(update_state)(json=true)
update_component_state_handler = RequestHandler.trans_cmd(update_component_state)(json=true)
state_handler = RequestHandler.trans_function(state_data)(name='state', sync=true, prefix=Full())
rpc_handlers_handler = RequestHandler.trans_function(rpc_handlers)(internal=true, sync=true, prefix=Full())
poll_handler = RequestHandler.trans_cmd(poll)(prefix=Full())
append_python_path_handler = RequestHandler.trans_function(append_python_path)(prefix=Full())
show_python_path_handler = RequestHandler.trans_function(show_python_path)(prefix=Full())
enable_components_handler = RequestHandler.trans_cmd(enable_components)(prefix=Full())
map_handler = RequestHandler.trans_function(mapping)(name='map', prefix=Full())
internal_init_handler = RequestHandler.trans_function(internal_init)(prefix=Full())


internal: Component = Component.cons(
    'internal',
    request_handlers=List(
        program_log_handler,
        set_log_level_handler,
        update_state_handler,
        update_component_state_handler,
        state_handler,
        rpc_handlers_handler,
        poll_handler,
        append_python_path_handler,
        show_python_path_handler,
        enable_components_handler,
        map_handler,
        internal_init_handler,
    ),
)


__all__ = ('internal',)



import os
import abc
import pty
import json
import time
import asyncio
import subprocess
from functools import wraps
from threading import Thread
from datetime import datetime
from contextlib import contextmanager
from typing import Any, Callable, Generic, TypeVar

import neovim
from neovim.api import Nvim

from amino import List, Either, __, env, Path, Lists, Map, do, Do, Nil, _, Maybe, Just, Nothing
from amino.lazy import lazy
from amino.test import fixture_path, temp_dir
from amino.test.path import base_dir, pkg_dir
from amino.test.spec import IntegrationSpecBase as AminoIntegrationSpecBase, default_timeout
from amino.util.string import camelcase
from amino.json import decode_json
from amino.json.data import JsonError

import ribosome
from ribosome.logging import Logging
from ribosome.nvim.api.data import NvimApi
from ribosome.config.settings import Settings
from ribosome.nvim.api.data import NativeNvimApi
from ribosome.nvim.api.option import option_cat
from ribosome.nvim.api.variable import variable_set_prefixed
from ribosome.nvim.api.function import define_function, nvim_call_function
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.command import nvim_command


def wait_for(cond: Callable[[], bool], timeout: float=None, intval: float=0.1) -> bool:
    to = default_timeout if timeout is None else timeout
    start = datetime.now()
    while (not cond() and (datetime.now() - start).total_seconds() < to):
        time.sleep(intval)
    return cond()


class IntegrationSpecBase(AminoIntegrationSpecBase):

    def setup(self) -> None:
        AminoIntegrationSpecBase.setup(self)

    def teardown(self) -> None:
        AminoIntegrationSpecBase.teardown(self)


class VimIntegrationSpecI(abc.ABC):

    @abc.abstractproperty
    def _log_out(self) -> List[str]:
        ...


def format_json_cmd(args: List[str], data: dict) -> str:
    j = json.dumps(data)
    return f'{args.join_tokens} {j}'


def json_cmd(cmd: str, *args: str, **data: str) -> NvimIO[str]:
    return nvim_command(cmd, format_json_cmd(Lists.wrap(args), data), verbose=True)


class VimIntegrationSpec(VimIntegrationSpecI, IntegrationSpecBase, Logging):

    def __init__(self) -> None:
        IntegrationSpecBase.__init__(self)
        self.tmux_nvim = 'RIBOSOME_TMUX_SPEC' in env
        self.tmux_nvim_external = False
        self.tmux_pane_id = None
        self.keep_tmux_pane = False
        self.vimlog = temp_dir('log') / 'vim'
        self.nvim_cmdline = List('nvim', '-V{}'.format(self.vimlog), '-n', '-u', 'NONE')
        self.log_format = '{levelname} {name}:{message}'
        self.subproc = None
        self.ribo_tmux = None
        self._debug = 'RIBOSOME_DEVELOPMENT' in env

    def setup(self) -> None:
        IntegrationSpecBase.setup(self)
        env['RIBOSOME_SPEC'] = 1
        env['AMINO_DEVELOPMENT'] = 1
        env['RIBOSOME_PKG_DIR'] = str(pkg_dir())
        self.logfile = temp_dir('log') / self.__class__.__name__
        self.logfile.touch()
        os.environ['RIBOSOME_LOG_FILE'] = str(self.logfile)
        os.environ['RIBOSOME_FILE_LOG_FMT'] = self.log_format
        self._pre_start_neovim()
        self.start_neovim()
        self._post_start_neovim()
        self._pre_start()
        self._start_plugin()
        self._post_start()

    def start_neovim(self) -> None:
        asyncio.get_child_watcher().attach_loop(asyncio.get_event_loop())
        if self.tmux_nvim:
            self.start_neovim_tmux_pane()
        elif self.tmux_nvim_external:
            self.start_neovim_tmux_bg()
        else:
            self.start_neovim_embedded()
        self.vim = self.create_nvim_api(self.neovim)

    @property
    def python_path(self) -> str:
        global_path = env['PYTHONPATH'] | ''
        return '{}:{}'.format(self.project_path, global_path)

    @property
    def vim_proc_env(self) -> Map:
        return Map(
            NVIM_LISTEN_ADDRESS=self.nvim_socket,
            PYTHONPATH=self.python_path,
            RIBOSOME_LOG_FILE=self.logfile,
            RIBOSOME_FILE_LOG_FMT=f'\'{self.log_format}\'',
            AMINO_DEVELOPMENT='1',
        )

    def start_neovim_embedded(self) -> None:
        ''' start an embedded vim session that loads no init.vim.
        **self.vimlog** is set as log file. aside from being convenient, this is crucially necessary, as the first use
        of the session will block if stdout is used for output.
        '''
        argv = self.nvim_cmdline.cat('--embed')
        self.neovim = neovim.attach('child', argv=argv)

    @lazy
    def nvim_socket(self) -> str:
        return str(temp_dir('nvim_sockets') / List.random_string())

    @property
    def project_path(self) -> str:
        return str(base_dir().parent)

    def connect_tmux(self, external: bool) -> Any:
        try:
            from chiasma.test.tmux_spec import tmux_spec_socket
            from chiasma.tmux import Tmux
        except ImportError:
            raise Exception('install chiasma to run nvim in a tmux pane')
        else:
            return Tmux.cons(socket=tmux_spec_socket if external else None)

    @property
    def tmux_window_external(self) -> Any:
        self.ribo_tmux = self.connect_tmux(True)
        from chiasma.io.compute import TmuxIO
        from chiasma.commands.window import windows
        @do(TmuxIO)
        def run() -> Do:
            ws = yield windows()
            yield TmuxIO.from_maybe(ws.head, 'tmux contains no windows')
        return run().unsafe(self.ribo_tmux)

    @property
    def tmux_window_local(self) -> Any:
        session = next(s for s in server.sessions if int(s['session_attached']) >= 1)
        return session.attached_window

    @lazy
    def tmux_window(self) -> Any:
        return self.tmux_window_external if self.tmux_nvim_external else self.tmux_window_local

    def start_neovim_tmux_bg(self) -> None:
        conf = fixture_path('conf', 'tmux.conf')
        args = ['tmux', '-L', self.tmux_socket, '-f', str(conf)]
        master, slave = pty.openpty()
        self.subproc = subprocess.Popen(args, stdout=slave, stdin=slave, stderr=slave)
        env_args = self.vim_proc_env.map2(lambda k, v: f'{k}={v}').cons('env')
        cmd = env_args + self.nvim_cmdline
        from chiasma.commands.pane import send_keys
        send_keys(0, List(cmd.join_tokens)).unsafe(self.ribo_tmux)
        wait_for(Path(self.nvim_socket).is_socket)

    def start_neovim_tmux_pane(self) -> None:
        self.ribo_tmux = self.connect_tmux(False)
        env_args = self.vim_proc_env.map2(lambda k, v: f'{k}={v}').cons('env')
        cmd = env_args + self.nvim_cmdline
        pid = os.getpid()
        from chiasma.io.compute import TmuxIO
        from chiasma.command import simple_tmux_cmd_attrs
        from psutil import Process
        @do(TmuxIO)
        def run() -> Do:
            ps = yield simple_tmux_cmd_attrs('list-panes', Nil, List('pane_pid', 'window_id'))
            def match(pane: Map[str, str]) -> bool:
                return Lists.wrap(Process(int(pane['pane_pid'])).children(recursive=True)).map(_.pid).contains(pid)
            current_pane = yield TmuxIO.from_maybe(ps.find(match), 'vim pane not found')
            window_id = current_pane['window_id']
            pane = yield simple_tmux_cmd_attrs('split-window', List('-t', window_id, '-d', '-P') + cmd, List('pane_id'))
            return pane[0]['pane_id'][1:]
        self.tmux_pane_id = run().unsafe(self.ribo_tmux)
        wait_for(Path(self.nvim_socket).is_socket)
        self.neovim = neovim.attach('socket', path=self.nvim_socket)
        self.neovim.command('python3 sys.path.insert(0, \'{}\')'.format(self.python_path))

    def _cleanup_tmux(self) -> None:
        if self.tmux_pane_id is not None and not self.keep_tmux_pane:
            from chiasma.commands.pane import close_pane_id
            close_pane_id(self.tmux_pane_id).unsafe(self.ribo_tmux)

    def create_nvim_api(self, vim: Nvim) -> NvimApi:
        return NativeNvimApi(self.plugin_name(), vim._session)

    @abc.abstractmethod
    def plugin_name(self) -> str:
        ...

    def plugin_short_name(self) -> str:
        return self.plugin_name()

    def full_cmd_prefix(self) -> str:
        return camelcase(self.plugin_name())

    def short_cmd_prefix(self) -> str:
        return camelcase(self.plugin_short_name())

    def teardown(self) -> None:
        IntegrationSpecBase.teardown(self)
        if self._debug:
            self._log_out.foreach(self.log.info)
        if self.tmux_nvim:
            self._cleanup_tmux()
        if self.subproc is not None:
            self.subproc.kill()
        if self.tmux_nvim_external:
            from chiasma.io.compute import TmuxIO
            TmuxIO.write('kill-server').unsafe(self.ribo_tmux)

    def _pre_start_neovim(self) -> None:
        pass

    def _post_start_neovim(self) -> None:
        rtp = fixture_path('config', 'rtp')
        option_cat('runtimepath', List(rtp)).unsafe(self.vim)
        variable_set_prefixed('debug', True)

    def _pre_start(self) -> None:
        pass

    def _start_plugin(self) -> None:
        pass

    def _post_start(self) -> None:
        pass

    @property
    def _log_out(self) -> List[str]:
        return List.wrap(self.logfile.read_text().splitlines())

    def _cmd(self, cmd: str, args: List[str]) -> str:
        return args.cons(cmd).join_tokens

    def cmd(self, cmd: str, *args: str) -> NvimIO[None]:
        return nvim_command(cmd, *args, verbose=True)

    def json_cmd(self, cmd: str, *args: str, **data: Any) -> NvimIO[None]:
        return json_cmd(cmd, *args, **data)

    @property
    def content(self) -> List[str]:
        return self.vim.buffer.content

    def program_log(self) -> Either[str, List['str']]:
        return self.vim.call(f'{self.full_cmd_prefix()}TransLog') // decode_json

    @property
    def state(self) -> Any:
        def error(err: JsonError) -> None:
            self.log.error(f'{err.error}: {err.data}')
            raise err.exception
        response = self.vim.call(f'{self.full_cmd_prefix()}State').get_or_raise()
        return decode_json(response).value_or(error)


def main_looped(fun):
    @wraps(fun)
    def wrapper(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        asyncio.get_child_watcher().attach_loop(loop)
        done = asyncio.Future(loop=loop)
        exc = None
        ret = None
        def runner():
            nonlocal ret
            local_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(local_loop)
            try:
                ret = fun(self)
            except Exception as e:
                nonlocal exc
                exc = e
            finally:
                loop.call_soon_threadsafe(lambda: done.set_result(True))
                local_loop.close()
        Thread(target=runner).start()
        loop.run_until_complete(done)
        loop.close()
        if exc is not None:
            raise exc
        return ret
    return wrapper


@contextmanager
def _nop_main_loop(self):
    yield


def _mock_async(self, f):
    ret = f(self)
    return ret


def _mock_proxy(self):
    return self


class ExternalIntegrationSpec(VimIntegrationSpec):

    def __init__(self) -> None:
        VimIntegrationSpec.__init__(self)
        self._report_expensive = False

    def _pre_start_neovim(self):
        super()._pre_start_neovim()
        ribosome.in_vim = False
        # NvimApi.async = _mock_async
        # NvimApi.main_event_loop = _nop_main_loop
        # NvimApi.proxy = property(_mock_proxy)
        # NvimApi.clean = lambda self: True

    def _post_start_neovim(self):
        cls = self.plugin_class.get_or_raise()
        self.plugin = cls(self.neovim)

    def _await(self):
        if self.root is not None:
            self.root.await_state()

    @property
    def root(self):
        return self.plugin.state()

    def teardown(self):
        VimIntegrationSpec.teardown(self)
        if self.root is not None:
            self.root.stop(shutdown=False)
            if self._debug and self._report_expensive:
                self._report()

    def _report(self):
        self.root.sub.cat(self.root) % __.report()


S = TypeVar('S', bound=Settings)
D = TypeVar('D')


class AutoPluginIntegrationSpec(Generic[S, D], VimIntegrationSpec):

    def setup(self) -> None:
        self.log_format = '{message}'
        del env['AMINO_COMPILE_COCO']
        super().setup()

    def module(self) -> str:
        return self.plugin_name()

    @property
    def autostart_plugin(self) -> bool:
        return True

    def _post_start_neovim(self) -> None:
        super()._post_start_neovim()
        if self.autostart_plugin:
            self.start_plugin()

    def start_plugin(self) -> None:
        stderr_handler_name = 'RibosomeSpecStderr'
        stderr_handler_body = '''
        let err = substitute(join(a:data, '\\r'), '"', '\\"', 'g')
        python3 import amino
        python3 from ribosome.logging import ribosome_envvar_file_logging
        python3 ribosome_envvar_file_logging()
        execute 'python3 amino.amino_log.error(f"""error starting rpc job on channel ' . a:id . ':\\r' . err . '""")'
        '''
        define_function(stderr_handler_name, List('id', 'data', 'event'), stderr_handler_body).unsafe(self.vim)
        cmd = f'from ribosome.host import start_module; start_module({self.module()!r})'
        args = ['python3', '-c', cmd]
        opts = dict(rpc=True, on_stderr=stderr_handler_name)
        nvim_call_function('jobstart', args, opts).unsafe(self.vim)

    def send_json(self, data: str) -> None:
        self.vim.call(f'{self.plugin_name()}Send', data)

    def _pre_start(self) -> None:
        if self.autostart_plugin:
            self.pvar_becomes('started', True)


__all__ = ('VimIntegrationSpec', 'ExternalIntegrationSpec', 'AutoPluginIntegrationSpec')



from typing import TypeVar, Callable, Any, Tuple
from concurrent.futures import wait, ThreadPoolExecutor

from neovim.msgpack_rpc.event_loop.base import BaseEventLoop

from amino import _, __, IO, Lists, Either, List, L, Nil
from amino.do import do, Do
from amino.case import Case
from amino.util.exception import format_exception
from amino.io import IOException
from amino.util.string import decode
from amino.state import EitherState

from ribosome.nvim.io.compute import NvimIO
from ribosome.logging import ribo_log
from ribosome.config.config import Config
from ribosome.nvim.io.state import NS
from ribosome.data.plugin_state import PluginState
# from ribosome.trans.action import LogMessage, Info, Error
from ribosome.compute.program import Program
from ribosome.config.settings import Settings
from ribosome import NvimApi
from ribosome.nvim.io.data import NResult, NSuccess, NError, NFatal
from ribosome.nvim.io.api import N
from ribosome.request.job import RequestJob
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.args import ParamsSpec
from ribosome.request.handler.arg_parser import ArgParser, JsonArgParser, TokenArgParser
from ribosome.compute.run import run_prog
from ribosome.data.plugin_state_holder import PluginStateHolder

Loop = TypeVar('Loop', bound=BaseEventLoop)
D = TypeVar('D')
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C', bound=Config)
R = TypeVar('R')
RDP = TypeVar('RDP')
S = TypeVar('S', bound=Settings)
CC = TypeVar('CC')


# class execute_io(Case, alg=DIO):

#     def iodio(self, io: IODIO[A]) -> NS[PluginState[S, D, CC], TransComplete]:
#         return NS.from_io(io.io)

#     def gather_i_os_dio(self, io: GatherIOsDIO[A]) -> NS[PluginState[S, D, CC], TransComplete]:
#         def gather() -> R:
#             gio = io.io
#             return gather_ios(gio.ios, gio.timeout)
#         return NS.from_io(IO.delay(gather))

#     def gather_subprocs_dio(self, io: GatherSubprocsDIO[A, TransComplete]) -> NS[PluginState[S, D, CC], TransComplete]:
#         ribo_log.debug(f'gathering {io}')
#         def gather() -> TransComplete:
#             gio = io.io
#             popens = gio.procs.map(__.execute(gio.timeout))
#             return gather_ios(popens, gio.timeout)
#         return NS.from_io(IO.delay(gather))

#     def nvim_iodio(self, io: NvimIODIO[A]) -> NS[PluginState[S, D, CC], TransComplete]:
#         return NS.lift(io.io)


# class program_log(Case, alg=LogMessage):

#     def info(self, msg: Info) -> NS[D, None]:
#         return NS.delay(lambda v: ribo_log.info(msg.message))

#     def error(self, msg: Error) -> NS[D, None]:
#         return NS.delay(lambda v: ribo_log.error(msg))


def arg_parser(handler: RequestHandler, params_spec: ParamsSpec) -> ArgParser:
    tpe = JsonArgParser if handler.json else TokenArgParser
    return tpe(params_spec)


def parse_args(handler: RequestHandler, args: List[Any]) -> NS[D, List[Any]]:
    return arg_parser(handler, handler.program.params_spec).parse(args)


def log_trans(trans: Program) -> NS[PluginState[S, D, CC], None]:
    return NS.pure(None) if trans.name in ('program_log', 'pure') else NS.modify(__.log_trans(trans.name))


@do(NS)
def run_request_handler(handler: RequestHandler, args: List[Any]) -> Do:
    parsed_args = yield NS.from_either(parse_args(handler, args))
    yield log_trans(handler.program)
    yield run_prog(handler.program, parsed_args)


@do(NS)
def traverse_programs(programs: List[Program], args: List[Any]) -> Do:
    yield programs.traverse(lambda a: run_request_handler(a, args), NS)


# TODO use from request_handler, since there is no concurrency handling anywhere else anymore.
@do(NvimIO[A])
def exclusive(holder: PluginStateHolder, f: Callable[[], NvimIO[Tuple[PluginState, A]]], desc: str) -> Do:
    '''this is the central unsafe function, using a lock and updating the state in `holder` in-place.
    '''
    yield holder.acquire()
    ribo_log.debug2(lambda: f'exclusive: {desc}')
    state, response = yield N.ensure_failure(f(), holder.release)
    yield N.delay(lambda v: holder.update(state))
    yield holder.release()
    ribo_log.debug2(lambda: f'release: {desc}')
    yield N.pure(response)


def exclusive_program(holder: PluginStateHolder, program: Program, args: List[Any], desc: str) -> NvimIO[R]:
    return exclusive(
        holder,
        lambda: run_request_handler(program, args).run(holder.state),
        desc
    )


def execute(state: PluginStateHolder[D], program: Program, args: List[Any]) -> NvimIO[Any]:
    return exclusive_program(state, program, args, program.name)


def regular_programs(name: str) -> EitherState[RequestJob, List[Program]]:
    return EitherState.inspect_f(lambda job: job.programs.lift(name).to_either(f'no program for {name}'))


def special_programs_sync(parts: List[str]) -> EitherState[RequestJob, List[Program]]:
    return regular_programs(parts.mk_string(':'))


def special_programs(head: str, tail: List[str]) -> EitherState[RequestJob, List[Program]]:
    return (
        special_programs_sync(tail)
        if head == 'sync' else
        regular_programs(tail.cons(head).mk_string(':'))
    )


@do(EitherState[RequestJob, List[Program]])
def job_programs() -> Do:
    name = yield EitherState.inspect(_.name)
    parts = Lists.split(name, ':')
    yield parts.detach_head.map2(special_programs) | (lambda: regular_programs(name))


@do(NvimIO[List[Any]])
def execute_request_job(job: RequestJob) -> Do:
    programs = yield N.e(job_programs().run_a(job))
    result = yield programs.traverse(L(execute)(job.state, _, job.args), NvimIO)
    ribo_log.debug(f'async job {job.name} completed')
    yield N.from_io(job.state.request_complete())
    yield N.pure(result)


class request_result(Case, alg=NResult):

    def __init__(self, job: RequestJob) -> None:
        self.job = job

    def n_success(self, result: NSuccess[List[Any]]) -> Any:
        desc = self.job.desc
        def multiple_results() -> int:
            ribo_log.error(f'multiple request handlers for {desc}')
            return 4
        def sync_result(head: Any, tail: List[Any]) -> Any:
            return head if tail.empty else multiple_results()
        def empty_result() -> int:
            ribo_log.error(f'no result in {desc}')
            return 3
        return result.value.detach_head.map2(sync_result) | empty_result if self.job.sync else 0

    def n_error(self, result: NError[List[Any]]) -> Any:
        ribo_log.error(result.error)
        return 1

    def n_fatal(self, result: NFatal[List[Any]]) -> Any:
        exc = result.exception
        tb = format_exception(exc).join_lines
        desc = self.job.desc
        ribo_log.error(f'fatal error in {desc}')
        ribo_log.debug(f'{desc} failed:\n{tb}')
        return 2


def request_job(state: PluginStateHolder[D], name: str, args: tuple, sync: bool) -> RequestJob:
    decoded_args = decode(args)
    fun_args = decoded_args.head | Nil
    bang = decoded_args.lift(1).contains(1)
    return RequestJob(state, decode(name), fun_args, sync, bang)


def execute_request(vim: NvimApi, state: PluginStateHolder[D], name: str, args: tuple, sync: bool) -> Any:
    sync_prefix = '' if sync else 'a'
    job = request_job(state, name, args, sync)
    ribo_log.debug(f'dispatching {sync_prefix}sync request: {job.name}({job.args})')
    result = request_result(job)(execute_request_job(job).result(vim))
    if sync:
        ribo_log.debug(f'request `{job.name}` completed: {result}')
    return decode(result)


__all__ = ('execute_request_job', 'execute', 'traverse_programs', 'compute_program')



from amino import Map, List, do, Do

from ribosome.compute.api import prog
from ribosome.config.component import Component
from ribosome.request.handler.prefix import Plain
from ribosome.request.handler.handler import RequestHandler
from ribosome.config.config import Config, NoData
from ribosome.nvim.api.variable import variable_set
from ribosome.nvim.io.state import NS

val = 71


@prog.unit
def trans1() -> NS[NoData, None]:
    return NS.lift(variable_set('autocmd_success', val))


core: Component = Component.cons('core')


config: Config = Config.cons(
    name='plug',
    components=Map(core=core),
    request_handlers=List(
        RequestHandler.trans_autocmd(trans1)('vim_resized', prefix=Plain()),
    ),
)

__all__ = ('config',)



import abc
import queue
import logging
from typing import TypeVar, Generic, Optional
from threading import Lock

import greenlet

from amino import Boolean, Maybe, Nothing, Try, do, Do, IO, Dat

from ribosome.nvim.io.compute import NvimIO, lift_n_result
from ribosome.logging import Logging
from ribosome.nvim.io.api import N
from ribosome.nvim.io.data import NResult
from ribosome.data.plugin_state import PluginState
from ribosome.config.settings import Settings

A = TypeVar('A')
CC = TypeVar('CC')
D = TypeVar('D')
S = TypeVar('S', bound=Settings)


class PluginStateHolder(Generic[D], Dat['PluginStateHolder[D]'], Logging):

    @staticmethod
    def concurrent(state: PluginState[S, D, CC], log_handler: logging.Handler=None) -> 'PluginStateHolder[D]':
        return ConcurrentPluginStateHolder(state, Maybe.check(log_handler), Lock(), False)

    @staticmethod
    def strict(state: PluginState[S, D, CC], log_handler: logging.Handler=None) -> 'PluginStateHolder[D]':
        return StrictPluginStateHolder(state, log_handler)

    def __init__(self, state: PluginState[S, D, CC], log_handler: Maybe[logging.Handler]=Nothing) -> None:
        self.state = state
        self.log_handler = log_handler

    def update(self, state: PluginState[S, D, CC]) -> None:
        self.state = state

    @abc.abstractmethod
    def acquire(self) -> NvimIO[None]:
        ...

    @abc.abstractmethod
    def release(self, result: Optional[NResult[A]]=None) -> NvimIO[None]:
        ...

    @abc.abstractmethod
    def request_complete(self) -> IO[None]:
        ...


class ConcurrentPluginStateHolder(Generic[D], PluginStateHolder[D]):

    def __init__(
            self,
            state: PluginState[S, D, CC],
            log_handler: Maybe[logging.Handler],
            lock: Lock,
            running: Boolean,
    ) -> None:
        super().__init__(state, log_handler)
        self.lock = lock
        self.running = running
        self.waiting_greenlets = queue.Queue()

    def _enqueue_greenlet(self) -> None:
        self.waiting_greenlets.put(greenlet.getcurrent())

    def _pop_greenlet(self, error: Exception=None) -> None:
        if error:
            self.log.caught_exception(f'switching to parent greenlet in `acquire`', error)
        gr = self.waiting_greenlets.get()
        self.waiting_greenlets.task_done()
        return gr

    @do(NvimIO[None])
    def acquire(self) -> Do:
        '''acquire the state lock that prevents multiple programs from updating the state asynchronously.
        If the lock is already acquired, an async request is currently executing while another (sync or async) has been
        requested. In order not to block on requests to vim from the running request, the greenlet that was started by
        the vim session must be suspended, giving control back to the running async request at the point where the vim
        request was made.
        '''
        def switch() -> None:
            self.log.debug(f'acquire: switching to running request')
            self._enqueue_greenlet()
            Try(lambda: greenlet.getcurrent().parent.switch()).leffect(self._pop_greenlet)
        if self.running:
            yield N.simple(switch)
        yield N.simple(self.lock.acquire)
        yield N.simple(setattr, self, 'running', True)

    @do(NvimIO[None])
    def release(self, result: Optional[NResult[A]]=None) -> Do:
        yield N.simple(setattr, self, 'running', False)
        yield N.simple(Try, self.lock.release)
        if result:
            self.log.debug(f'released lock due to error: {result}')
            yield lift_n_result.match(result)

    @do(IO[None])
    def request_complete(self) -> Do:
        '''switch back to the request that was suspended in `acquire` while the request executing this method was
        running.
        '''
        if not self.waiting_greenlets.empty():
            self.log.debug('release: switching to waiting request')
            gr = yield IO.delay(self._pop_greenlet)
            yield IO.delay(gr.switch)
        yield IO.pure(None)


class StrictPluginStateHolder(Generic[D], PluginStateHolder[D]):
    pass

    def acquire(self) -> NvimIO[None]:
        return N.pure(None)

    def release(self, result: Optional[NResult[A]]=None) -> NvimIO[None]:
        return N.pure(None)

    def request_complete(self) -> IO[None]:
        return IO.pure(None)


__all__ = ('PluginStateHolder', 'ConcurrentPluginStateHolder', 'StrictPluginStateHolder')



import logging
from typing import TypeVar, Generic, Callable, Any, Type
from uuid import UUID

from amino import Map, List, Nil, Either, _, Maybe, Nothing, __, L
from amino.dat import Dat
from amino.util.string import camelcase

from ribosome.config.component import Component, Components, NoComponentData, ComponentConfig
from ribosome.nvim.io.state import NS
from ribosome.compute.program import Program
from ribosome.request.rpc import RpcHandlerSpec, DefinedHandler
from ribosome.config.settings import Settings
from ribosome.config.resources import Resources
from ribosome.request.handler.handler import RequestHandler, RequestHandlers
from ribosome.request.handler.method import RpcMethod
from ribosome.config.basic_config import BasicConfig
from ribosome.compute.output import ProgIO
from ribosome.compute.prog import Prog
from ribosome.compute.interpret import interpret_io, no_interpreter

A = TypeVar('A')
C = TypeVar('C')
CC = TypeVar('CC')
D = TypeVar('D')
DIO = TypeVar('DIO')
Meth = TypeVar('Meth', bound=RpcMethod)
P = TypeVar('P')
S = TypeVar('S', bound=Settings)
Programs = Map[str, Program]


def handler_spec(handler: RequestHandler[Meth, P], name: str, prefix: str) -> RpcHandlerSpec:
    return RpcHandlerSpec.cons(
        handler.method.spec_type,
        handler.sync,
        handler.vim_cmd_name(name, prefix),
        handler.options,
        handler.method_str,
        True,
    )


# FIXME in order to allow transparent updating of components, the data and config part must be separated
# FIXME using state type as a key won't work if multiple comps have the same type
class PluginState(Generic[S, D, CC], Dat['PluginState[S, D, CC]']):

    @staticmethod
    def cons(
            basic: BasicConfig,
            comp: ComponentConfig,
            request_handlers: RequestHandlers,
            data: D,
            components: List[Component],
            init: Program,
            program_log: List[str]=Nil,
            logger: Program[None]=None,
            log_handler: logging.Handler=None,
            component_data: Map[type, Any]=Map(),
            active_mappings: Map[UUID, Program]=Map(),
            io_executor: Callable[[DIO], NS['PluginState[S, D, CC]', Any]]=None,
            rpc_handlers: List[DefinedHandler]=Nil,
            programs: Programs=Map(),
            io_interpreter: Callable[[ProgIO], Prog]=None,
            custom_io: Callable[[Any], Prog[A]]=None,
    ) -> 'PluginState':
        components = Components.cons(components)
        return PluginState(
            basic,
            comp,
            request_handlers,
            data,
            components,
            init,
            program_log,
            log_handler,
            component_data,
            active_mappings,
            rpc_handlers,
            Maybe.optional(io_executor),
            programs,
            io_interpreter or interpret_io(custom_io or no_interpreter, Maybe.optional(logger)),
        )

    def __init__(
            self,
            basic: BasicConfig,
            comp: ComponentConfig,
            request_handlers: RequestHandlers,
            data: D,
            components: Components,
            init: Program,
            program_log: List[str],
            log_handler: Maybe[logging.Handler],
            component_data: Map[type, Any],
            active_mappings: Map[UUID, Program],
            io_executor: Maybe[Callable[[DIO], NS['PluginState[S, D, CC]', Any]]],
            rpc_handlers: List[DefinedHandler],
            programs: Programs,
            io_interpreter: Callable[[ProgIO], Prog],
    ) -> None:
        self.basic = basic
        self.comp = comp
        self.data = data
        self.components = components
        self.init = init
        self.program_log = program_log
        self.log_handler = log_handler
        self.component_data = component_data
        self.active_mappings = active_mappings
        self.io_executor = io_executor
        self.rpc_handlers = rpc_handlers
        self.programs = programs
        self.request_handlers = request_handlers
        self.io_interpreter = io_interpreter

    def update(self, data: D) -> 'PluginState[S, D, CC]':
        return self.copy(data=data)

    def log_trans(self, trans: str) -> 'PluginState[S, D, CC]':
        return self.append1.program_log(trans)

    def component(self, name: str) -> Either[str, Component]:
        return self.components.by_name(name)

    def ctor_by_type(self, tpe: Type[C]) -> Callable[[], C]:
        return self.components.by_type(tpe) / _.state_ctor | (lambda: NoComponentData)

    def data_by_type(self, tpe: Type[C]) -> C:
        return self.component_data.lift(tpe) | (lambda: self.ctor_by_type(tpe)())

    def data_for(self, component: Component) -> Any:
        return self.component_data.lift(component.state_type) | component.state_ctor

    def data_by_name(self, name: str) -> Either[str, Any]:
        return self.component(name) / self.data_for

    def update_component_data(self, new: C) -> 'PluginState[S, D, CC]':
        return self.mod.component_data(__.cat((type(new), new)))

    def modify_component_data(self, name: str, mod: Callable[[Any], Any]) -> 'PluginState[S, D, CC]':
        comp = self.data_by_name(name)
        return comp / mod / self.update_component_data | self

    def resources_with(self, data: A) -> Resources[S, A, CC]:
        return Resources(data, self.basic.settings, self.components)

    @property
    def resources(self) -> Resources[S, D, CC]:
        return self.resources_with(self.data)

    @property
    def camelcase_name(self) -> str:
        return camelcase(self.basic.name)

    @property
    def specs(self) -> List[RpcHandlerSpec]:
        return self.programs.v.join / L(handler_spec)(_, self.basic.name, self.basic.prefix)

    @property
    def distinct_specs(self) -> List[RpcHandlerSpec]:
        return self.specs.distinct_by(_.rpc_method)


__all__ = ('PluginState',)



from typing import Callable, TypeVar, Generic, Optional, Type, Any

from amino import List, Nil, Map, Either, Right, do, Do, Maybe
from amino.dat import Dat
from amino.json.encoder import Encoder, json_object_with_type
from amino.json.data import JsonError, Json, JsonScalar
from amino.json.decoder import Decoder, decode

from ribosome.request.handler.handler import RequestHandler, RequestHandlers
from ribosome.config.settings import Settings
from ribosome.request.rpc import RpcHandlerSpec
from ribosome.config.component import Component
from ribosome.compute.program import Program
from ribosome.config.basic_config import NoData, BasicConfig
from ribosome.components.internal.config import internal

A = TypeVar('A')
D = TypeVar('D')
S = TypeVar('S', bound=Settings)
CC = TypeVar('CC')
G = TypeVar('G')


class Config(Generic[S, D, CC], Dat['Config[S, D, CC]']):

    @staticmethod
    def from_opt(data: Map) -> 'Config[S, D, CC]':
        return Config.cons(data.lift('name') | 'no name in json', data.lift('prefix') | None)

    @staticmethod
    def cons(
            name: str,
            prefix: Optional[str]=None,
            state_ctor: Callable[[], D]=None,
            components: Map[str, Component[D, Any, CC]]=Map(),
            settings: S=None,
            request_handlers: List[RequestHandler]=Nil,
            core_components: List[str]=Nil,
            default_components: List[str]=Nil,
            init: Program=None,
            internal_component: bool=True,
    ) -> 'Config[S, D, CC]':
        basic = BasicConfig.cons(
            name,
            prefix,
            state_ctor,
            settings,
            core_components,
            default_components,
            internal_component,
        )
        return Config(
            basic,
            components + ('internal', internal) if internal_component else components,
            RequestHandlers.cons(*request_handlers),
            Maybe.optional(init),
        )

    def __init__(
            self,
            basic: BasicConfig[S, D],
            components: Map[str, Component[D, Any, CC]],
            request_handlers: RequestHandlers,
            init: Maybe[Program],
    ) -> None:
        self.basic = basic
        self.components = components
        self.request_handlers = request_handlers
        self.init = init

    def _arg_desc(self) -> List[str]:
        return List(str(self.components), str(self.settings), str(self.request_handlers))

    def vim_cmd_name(self, handler: RequestHandler) -> str:
        return handler.vim_cmd_name(self.name, self.prefix)

    @property
    def rpc_specs(self) -> List[RpcHandlerSpec]:
        return self.request_handlers.rpc_specs(self.name, self.prefix)


class ConfigEncoder(Encoder[Config], tpe=Config):

    def encode(self, a: Config) -> Either[JsonError, Map]:
        return Right(json_object_with_type(Map(name=JsonScalar(a.name)), Config))


class ConfigDecoder(Decoder[Config], tpe=Config):

    @do(Either[JsonError, Config])
    def decode(self, tpe: Type[Config], data: Json) -> Do:
        f = yield data.field('name')
        name = yield decode(f)
        yield Right(Config.cons(name))


__all__ = ('Config', 'NoData', 'Resources')



from typing import TypeVar, Callable, Generic, Any, Type, Union

from amino import List, Either, _, Nil, Maybe, Boolean, __, Map
from amino.dat import Dat

from ribosome.request.handler.handler import RequestHandler, RequestHandlers
from ribosome.nvim.io.state import NS
from ribosome.data.mapping import Mappings

# FIXME typevar P is temporary
D = TypeVar('D')
P = TypeVar('P')
CC = TypeVar('CC')
CD = TypeVar('CD')


class NoComponentData(Dat['NoComponentData']):
    pass


class ComponentData(Generic[D, CD], Dat['ComponentData[D, CD]']):

    def __init__(self, main: D, comp: CD) -> None:
        self.main = main
        self.comp = comp


def comp_data() -> NS[ComponentData[D, CD], CD]:
    return NS.inspect(_.comp)


def infer_state_ctor(state_type: Union[None, Type[CD]]) -> Union[None, Callable[[], CD]]:
    return (
        None
        if state_type is None else
        getattr(state_type, 'cons', state_type)
    )


class Component(Generic[D, CD, CC], Dat['Component[D, CD, CC]']):

    @staticmethod
    def cons(
            name: str,
            request_handlers: List[RequestHandler]=Nil,
            trans_handlers: List[RequestHandler]=Nil,
            state_type: Type[CD]=None,
            state_ctor: Callable[[], CD]=None,
            config: CC=None,
            mappings: Mappings=None,
    ) -> 'Component[D, CD, CC]':
        return Component(
            name,
            RequestHandlers.cons(*request_handlers),
            RequestHandlers.cons(*trans_handlers),
            state_type or NoComponentData,
            state_ctor or infer_state_ctor(state_type) or NoComponentData,
            Maybe.check(config),
            mappings or Mappings.cons(),
        )

    def __init__(
            self,
            name: str,
            request_handlers: RequestHandlers,
            trans_handlers: RequestHandlers,
            state_type: Type[CD],
            state_ctor: Maybe[Callable[[D], CD]],
            config: Maybe[CC],
            mappings: Mappings,
    ) -> None:
        self.name = name
        self.request_handlers = request_handlers
        self.trans_handlers = trans_handlers
        self.state_type = state_type
        self.state_ctor = state_ctor
        self.config = config
        self.mappings = mappings

    def handler_by_name(self, name: str) -> Either[str, P]:
        return self.handlers.find(_.name == name).to_either(f'component `{self.name}` has no handler `name`')

    def contains(self, handler: P) -> Boolean:
        return self.request_handlers.trans_handlers.exists(_.fun == handler.fun)


class Components(Generic[D, CC], Dat['Components']):

    @staticmethod
    def cons(
            all: List[Component[D, Any, CC]]=Nil,
    ) -> 'Components[D, CC]':
        return Components(all)

    def __init__(self, all: List[Component[D, Any, CC]]) -> None:
        self.all = all

    def by_name(self, name: str) -> Either[str, Component[D, CD, CC]]:
        return self.all.find(_.name == name).to_either(f'no component named {name}')

    def by_type(self, tpe: type) -> Either[str, Component[D, CD, CC]]:
        return self.all.find(_.state_type == tpe).to_either(f'no component with state type {tpe}')

    @property
    def config(self) -> List[CC]:
        return self.all.collect(_.config)

    def for_handler(self, handler: P) -> Maybe[Component]:
        return self.all.find(__.contains(handler))


class ComponentConfig(Dat['ComponentConfig']):

    @staticmethod
    def cons(
            available: Map[str, Component[D, Any, CC]],
    ) -> 'ComponentConfig':
        return ComponentConfig(
            available,
        )

    def __init__(self, available: Map[str, Component[D, Any, CC]]) -> None:
        self.available = available


__all__ = ('Component', 'Components')



from typing import Any

from ribosome.config.component import Component

from amino import Map, Lists
from ribosome.request.handler.handler import RequestHandler
from ribosome.config.config import Config
from ribosome.compute.program import Program


default_config_name = 'spec'


def spec_config(*request_handlers: RequestHandler) -> Config:
    component = Component.cons(
        'main',
        request_handlers=Lists.wrap(request_handlers)
    )
    return Config.cons(
        name=default_config_name,
        prefix=default_config_name,
        components=Map(main=component),
    )


def single_trans_config(trans: Program, **kw: Any) -> Config:
    return spec_config(RequestHandler.trans_cmd(trans)(**kw))


__all__ = ('single_trans_config', 'spec_config')



from typing import Callable, Tuple, Any, TypeVar, Generic

from amino import Lists, Map, List, __, Either, Left, Right, do, Do, Nil

from kallikrein.expectable import Expectable
from amino.dat import Dat
from amino.lenses.lens import lens

from ribosome.host import init_state
from ribosome.request.execute import request_job, traverse_programs, run_request_handler
from ribosome.nvim.io.state import NS
from ribosome.config.config import Config
from ribosome.config.settings import Settings
from ribosome.config.component import ComponentData
from ribosome.nvim.api.data import NvimApi, StrictNvimApi
from ribosome.nvim.io.compute import NvimIO
from ribosome.data.plugin_state import PluginState
from ribosome.nvim.api.variable import variable_set
from ribosome.nvim.io.api import N
from ribosome.request.job import RequestJob
from ribosome.compute.program import Program
from ribosome.config.resources import Resources
from ribosome.data.plugin_state_holder import PluginStateHolder
from ribosome.compute.output import ProgOutput
from ribosome.compute.prog import Prog
from ribosome.test.klk import kn

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
S = TypeVar('S', bound=Settings)
CC = TypeVar('CC')
DIO = TypeVar('DIO')
R = TypeVar('R')
ProgInt = Callable[[ProgOutput], Callable[[R], Prog[B]]]
Handler = Callable[['StrictNvimApi', str, List[Any]], Either[str, Tuple[NvimApi, Any]]]


def api_info(vim: NvimApi, name: str, args: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    return Right((vim, (1, {})))


def rh_write(vim: NvimApi, name: str, args: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    return Right((vim, None))


@do(Either[str, Tuple[NvimApi, Any]])
def default_request_handler(vim: NvimApi, name: str, args: List[Any]) -> Do:
    handler = yield default_request_handlers.lift(name).to_either(f'no default request handler for {name}')
    yield handler(vim, name, args)


@do(Either[str, Tuple[NvimApi, Any]])
def pop_first(vim: NvimApi, name: str, args: List[Any]) -> Do:
    name1, args1 = yield args.detach_head.to_either(f'invalid command: {name}({args})')
    yield default_request_handler(vim, name1, args1)


default_request_handlers = Map({
    'silent': pop_first,
    'silent!': pop_first,
    'nvim_get_api_info': api_info,
    'nvim_command': rh_write,
    'nvim_out_write': rh_write,
    'nvim_call_function': pop_first,
})


@do(Either[str, Tuple[NvimApi, Any]])
def specific_handler(desc: str, handler: Handler, vim: StrictNvimApi, name: str, args: List[Any]) -> Do:
    name1, args1 = yield args.detach_head.to_either(f'empty {desc} args: {name}')
    yield handler(vim, name1, args1)


@do(Either[str, Tuple[NvimApi, Any]])
def function_handler(handler: Handler, vim: StrictNvimApi, name: str, args: List[Any]) -> Do:
    yield specific_handler('function call', handler, vim, name, args)


@do(Either[str, Tuple[NvimApi, Any]])
def command_handler(handler: Handler, vim: StrictNvimApi, name: str, args: List[Any]) -> Do:
    yield specific_handler('command', handler, vim, name, args)


class StrictRequestHandler:

    def __init__(
            self,
            extra: Handler,
            function_handler: Handler,
            command_handler: Handler,
    ) -> None:
        self.extra = extra
        self.function_handler = function_handler
        self.command_handler = command_handler

    def __call__(self, vim: NvimApi, name: str, args: List[Any]) -> Either[List[str], Tuple[NvimApi, Any]]:
        return (
            self.extra(vim, name, args)
            .lmap(List)
            .accum_error_lift(function_handler, self.function_handler, vim, name, args)
            .accum_error_lift(command_handler, self.command_handler, vim, name, args)
            .accum_error_lift(default_request_handler, vim, name, args)
        )


def no_handler(vim: NvimApi, name: str, args: List[Any]) -> Either[str, Tuple[NvimApi, Any]]:
    return Left(f'no handler for {name}')


class RequestHelperBuilder(Generic[S, D, CC], Dat['RequestHelperBuilder[S, D, CC]']):

    def __init__(
            self,
            config: Config[S, D, CC],
            comps: List[str],
            vars: Map[str, Any]=Map(),
            io_interpreter: ProgInt=None,
            logger: Program[None]=None,
    ) -> None:
        self.config = config
        self.comps = comps
        self.vars = vars
        self.io_interpreter = io_interpreter
        self.logger = logger

    def create(self, cons_vim: Callable[[dict], NvimApi]) -> 'RequestHelper[S, D, CC]':
        comps_var = Map({f'{self.config.basic.name}_components': self.comps})
        vim = cons_vim(self.vars ** comps_var)
        vim1, state = init_state(self.config, io_interpreter=self.io_interpreter, logger=self.logger).unsafe_run(vim)
        return RequestHelper(vim, self.config, state)

    def nvim(self, vim: NvimApi) -> 'RequestHelper[S, D, CC]':
        def cons_vim(vars: dict) -> NvimApi:
            for k, v in vars.items():
                variable_set(k, v).unsafe(vim)
            return vim
        return self.create(cons_vim)

    def strict(
            self,
            request_handler: Handler=no_handler,
            function_handler: Handler=no_handler,
            command_handler: Handler=no_handler,
    ) -> 'RequestHelper[S, D, CC]':
        def cons_vim(vars: Map[str, Any]) -> StrictNvimApi:
            handler = StrictRequestHandler(request_handler, function_handler, command_handler)
            return StrictNvimApi(self.config.basic.name, vars, handler, Nil)
        return self.create(cons_vim)


class RequestHelper(Generic[S, D, CC], Dat['RequestHelper[S, D, CC]']):

    @staticmethod
    def nvim(config: Config[S, D, CC], vim: NvimApi, *comps: str, **kw: Any) -> 'RequestHelper[S, D, CC]':
        return RequestHelper.cons(config, *comps, **kw).nvim(vim)

    @staticmethod
    def strict(
            config: Config,
            *comps: str,
            request_handler: Handler=no_handler,
            **kw: Any,
    ) -> 'RequestHelper':
        return RequestHelper.cons(config, *comps, **kw).strict(request_handler)

    @staticmethod
    def cons(config: Config[S, D, CC], *comps: str, **kw: Any) -> RequestHelperBuilder[S, D, CC]:
        return RequestHelperBuilder(config, Lists.wrap(comps), **kw)

    def __init__(self, vim: NvimApi, config: Config, state: PluginState[S, D, CC]) -> None:
        self.vim = vim
        self.config = config
        self.state = state

    @property
    def holder(self) -> PluginStateHolder[D]:
        return PluginStateHolder.strict(self.state)

    @property
    def settings(self) -> S:
        return self.state.basic.settings

    def request_job(self, name: str, args: tuple, sync: bool=True) -> Tuple[RequestJob, List[Program]]:
        job = request_job(self.holder, name, (args,), sync)
        return job, job.state.state.programs.lift(job.name).get_or_fail(f'no matching program for `{job.name}`')

    def traverse_programs(self, name: str, args: tuple=(), sync: bool=True) -> NS[PluginState[S, D, CC], Any]:
        job, program = self.request_job(name, args, sync)
        return traverse_programs(program, Lists.wrap(args))

    def run(self, name: str, args=(), sync: bool=True) -> NvimIO[Tuple[PluginState, Any]]:
        job, programs = self.request_job(name, args, sync)
        program = programs.head.get_or_fail('multiple programs')
        return (
            run_request_handler(program, Lists.wrap(args))
            .run(self.state)
        )

    def run_s(self, name: str, args=(), sync: bool=True) -> NvimIO[PluginState]:
        return self.run(name, args, sync).map2(lambda s, a: s)

    def run_a(self, name: str, args=(), sync: bool=True) -> NvimIO[Any]:
        return self.run(name, args, sync).map2(lambda s, a: a)

    def unsafe_run(self, name: str, args=(), sync: bool=True) -> Tuple[PluginState, Any]:
        return self.run(name, args=args, sync=sync).unsafe(self.vim)

    def unsafe_run_s(self, name: str, args=(), sync: bool=True) -> PluginState:
        return self.run_s(name, args=args, sync=sync).unsafe(self.vim)

    def unsafe_run_a(self, name: str, args=(), sync: bool=True) -> Any:
        return self.run_a(name, args=args, sync=sync).unsafe(self.vim)

    def update_data(self, **kw: Any) -> 'RequestHelper[S, D, CC]':
        return lens.state.data.modify(__.copy(**kw))(self)

    def update_component(self, name: str, **kw: Any) -> 'RequestHelper[S, D, CC]':
        return self.mod.state(__.modify_component_data(name, __.copy(**kw)))

    def component_res(self, data: C) -> Resources[S, C, CC]:
        return self.state.resources_with(ComponentData(self.state.data, data))

    def component_res_for(self, name: str) -> Resources[S, C, CC]:
        data = self.state.data_by_name(name).get_or_raise()
        return self.state.resources_with(ComponentData(self.state.data, data))

    def k(self, f: Callable[..., NvimIO[A]], *a: Any, **kw: Any) -> Expectable:
        return kn(self.vim, f, *a, **kw)


def request_helper(
        config: Config[S, D, CC],
        *comps: str,
        vars: dict=dict(),
        io_interpreter: ProgInt=None,
) -> NvimIO[RequestHelper[S, D, CC]]:
    return N.delay(lambda v: RequestHelper.nvim(config, v, *comps, vars=vars, io_interpreter=io_interpreter))


__all__ = ('DispatchHelper', 'request_helper')
